from __future__ import annotations

from typing import Optional
from enum import Enum


class IndexDescriptorType(Enum):
    ANY = "*"
    POSITIVE = "+"
    CONSTANT = ""


class IndexDescriptor():

    def __init__(self, vdtype: IndexDescriptorType, value: Optional[int] = None):  # pylint: disable=unsubscriptable-object
        if vdtype == IndexDescriptorType.CONSTANT and value is None:
            raise ValueError(
                "Value cannot be None for constant type descriptor!")
        if vdtype != IndexDescriptorType.CONSTANT and value is not None:
            raise ValueError(
                "Value must be None for non constant type descriptor!")
        self._vdtype = vdtype
        self._value = value

    @property
    def vdtype(self) -> IndexDescriptorType:
        return self._vdtype

    @property
    def converted_value(self) -> int:
        return self._value if self.vdtype == IndexDescriptorType.CONSTANT else 1

    def __str__(self):
        return f"{self.vdtype.value}{self._value if self._value is not None else ''}"

    def __repr__(self):
        return f"{self.converted_value}"


class IndexSet():

    def __init__(self, descriptors: tuple[IndexDescriptor]):
        if len(descriptors) == 0:
            raise ValueError("There must be at least one variable descriptor!")
        self._descriptors = descriptors

    @property
    def descriptors(self) -> tuple[IndexDescriptor]:
        return self._descriptors

    @property
    def converted_values(self) -> tuple[int]:
        return tuple(descriptor.converted_value for descriptor in self.descriptors)

    def is_lexico_positive(self) -> bool:
        for v in self.converted_values:
            if v > 0:
                return True
            if v < 0:
                return False
        return False

    def extract_positives(self) -> list[IndexSet]:
        expanded_sets = [self]
        # Expand * into 0 and +
        for i, desc in enumerate(self.descriptors):
            if desc.vdtype == IndexDescriptorType.ANY:
                new_sets = []
                for exp_set in expanded_sets:
                    exp_descs = list(exp_set.descriptors)
                    zero_descs = exp_descs[:i] + [IndexDescriptor(
                        IndexDescriptorType.CONSTANT, 0)] + exp_descs[i+1:]
                    plus_descs = exp_descs[:i] + [IndexDescriptor(
                        IndexDescriptorType.POSITIVE)] + exp_descs[i+1:]
                    zero_set = IndexSet(zero_descs)
                    plus_set = IndexSet(plus_descs)
                    if zero_set.is_lexico_positive():
                        new_sets.append(zero_set)
                    new_sets.append(plus_set)
                expanded_sets = new_sets
        # Filter out non-positive sets
        return list(filter(lambda s: s.is_lexico_positive(), expanded_sets))
    
    def to_converted(self) -> IndexSet:
        return IndexSet(list([
            IndexDescriptor(IndexDescriptorType.CONSTANT, value=v) for v in self.converted_values
        ]))

    def __str__(self):
        return f"({','.join([str(descriptor) for descriptor in self.descriptors])})"

    def __repr__(self):
        return f"({','.join([repr(descriptor) for descriptor in self.descriptors])})"

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(tuple([str(d) for d in self.descriptors]))


from abc import ABC, abstractmethod
from enum import auto, Enum

from collections import Counter

from opoly.expressions import (
    Expression,
    SingleExpression,
    VariableExpression,
    ConstantExpression,
    extract_variable_expressions
)


class StatementType(Enum):
    FOR_LOOP = auto()
    DECLARATION = auto()
    ASSIGNMENT = auto()


class Statement(ABC):

    def __init__(self, stype: StatementType):
        self._stype = stype

    @property
    def stype(self) -> StatementType:
        return self._stype

    @abstractmethod
    def stringify(self) -> str:
        pass

    def __str__(self):
        return self.stringify()

class SimpleStatement(Statement, ABC):

    def __init__(self, stype: StatementType):
        super().__init__(stype)

class AssignmentStatement(SimpleStatement):

    def __init__(self, left_term: VariableExpression, right_term: Expression):
        super().__init__(StatementType.ASSIGNMENT)
        self._left_term = left_term
        self._right_term = right_term

    @property
    def left_term(self) -> VariableExpression:
        return self._left_term

    @property
    def right_term(self) -> Expression:
        return self._right_term

    def stringify(self) -> str:
        return f"{self.left_term} = {self.right_term}"


class DeclarationStatement(SimpleStatement):
    def __init__(self, var_type: str, variable: VariableExpression, initialization: Expression = None):
        super().__init__(StatementType.DECLARATION)
        self._var_type = var_type
        self._variable = variable
        self._initialization = initialization

    @property
    def var_type(self) -> str:
        return self._var_type

    @property
    def variable(self) -> VariableExpression:
        return self._variable

    @property
    def initialization(self) -> Expression:
        return self._initialization

    def stringify(self) -> str:
        decl_str = f"{self.var_type} {str(self.variable)}"
        init_str = f" = {str(self.initialization)}" if self.initialization is not None else ""
        return f"{decl_str}{init_str}"


class CompoundStatement(Statement, ABC):

    def __init__(self, stype: StatementType, body: tuple[Statement]):
        super().__init__(stype)
        if len(body) == 0:
            raise ValueError("Compound statement body cannot be empty!")
        self._body = body

    @property
    def body(self) -> tuple[Statement]:
        return self._body

    @abstractmethod
    def stringify_head(self) -> str:
        pass

    def stringify(self) -> str:
        head = self.stringify_head()
        body = "\n".join([str(s) for s in self.body])
        return "\n".join([head, body])


class ForLoopStatement(CompoundStatement):

    def __init__(self,
                 body: tuple[Statement],
                 index: VariableExpression,
                 lowerbound: Expression,
                 upperbound: Expression,
                 step: Expression = ConstantExpression(1),
                 is_parallel: bool = False
                 ):
        super().__init__(StatementType.FOR_LOOP, body)
        self._index = index
        self._lowerbound = lowerbound
        self._upperbound = upperbound
        self._step = step
        self._is_parallel = is_parallel

    @property
    def index(self) -> VariableExpression:
        return self._index

    @property
    def lowerbound(self) -> Expression:
        return self._lowerbound

    @property
    def upperbound(self) -> Expression:
        return self._upperbound

    @property
    def step(self) -> Expression:
        return self._step
    
    @property
    def is_parallel(self) -> bool:
        return self._is_parallel

    def stringify_head(self) -> str:
        head = f"FOR {self.index} = {self.lowerbound}...{self.upperbound}"
        step = "" if isinstance(self.step, ConstantExpression) or self.step.value == 1 \
            else f" step {self.step}"
        return f"{head}{step}"


def divide_assignments(
    assignments: tuple[AssignmentStatement]
) -> (tuple[VariableExpression], tuple[VariableExpression]):
    generations = []
    uses = []
    for ass in assignments:
        generations.append(ass.left_term)
        uses.extend(extract_variable_expressions(ass.right_term))
    return (tuple(generations), tuple(uses))


def prune_expressions(
    generations: tuple[VariableExpression], uses: tuple[VariableExpression]
) -> (tuple[VariableExpression], tuple[VariableExpression]):
    gen_names = set(gen.name for gen in generations)
    use_names = set(use.name for use in uses)
    both_names = gen_names & use_names

    count_names = Counter(var.name for var in generations + uses)

    generations = filter(
        lambda g: g.name in gen_names and count_names[g.name] > 1, generations)
    uses = filter(lambda u: u.name in both_names, uses)
    return (tuple(generations), tuple(uses))



from pkg_resources import resource_filename, resource_dir
LAMPORT_SCHEDULER_PATH = resource_filename(__name__, "models/lamport_scheduler.mzn")
LAMPORT_ALLOCATOR_PATH = resource_filename(__name__, "models/lamport_allocator.mzn")
INCLUDE_FOLDER_PATH = resource_filename(__name__, "libraries/")


from opoly.expressions import Expression, VariableExpression, ConstantExpression
from opoly.statements import ForLoopStatement, DeclarationStatement, AssignmentStatement
from opoly.modules.generator import PseudoCodeGenerator, CCodeGenerator


class TestPseudoCodeGenerator():

    def test_declaration(self):
        stmt = DeclarationStatement(
            var_type="",
            variable=VariableExpression("i"),
            initialization=ConstantExpression(1)
        )
        code = PseudoCodeGenerator().generate(stmt)
        assert code == "VAR i = 1;"

    def test_assignment(self):
        stmt = AssignmentStatement(
            left_term=VariableExpression("i"),
            right_term=VariableExpression("j")
        )
        code = PseudoCodeGenerator().generate(stmt)
        assert code == "STM i = j;"

    def test_simple_for_loop(self):
        stmt = ForLoopStatement(
            body=[
                DeclarationStatement(
                    var_type="",
                    variable=VariableExpression("x"),
                    initialization=ConstantExpression(1)
                ),
                AssignmentStatement(
                    left_term=VariableExpression(
                        "a", [VariableExpression("i")]),
                    right_term=Expression([
                        VariableExpression("x"),
                        ConstantExpression(1)
                    ], ["+"])
                )
            ],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        code = PseudoCodeGenerator().generate(stmt)
        assert code == "FOR i FROM 1 TO N STEP 1 {\n    VAR x = 1;\n    STM a[i] = x + 1;\n}"

    def test_nested_for_loop(self):
        inner_loop = ForLoopStatement(
            body=[
                DeclarationStatement(
                    var_type="",
                    variable=VariableExpression("x"),
                    initialization=ConstantExpression(1)
                ),
                AssignmentStatement(
                    left_term=VariableExpression(
                        "a", [VariableExpression("j")]),
                    right_term=Expression([
                        VariableExpression("x"),
                        ConstantExpression(1)
                    ], ["+"])
                )
            ],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("M"),
            is_parallel=True
        )
        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        code = PseudoCodeGenerator().generate(outer_loop)
        assert code == "FOR i FROM 1 TO N STEP 1 {\n    FOR CONC j FROM 1 TO M STEP 1 {\n        VAR x = 1;\n        STM a[j] = x + 1;\n    }\n}"


class TestCCodeGenerator():

    def test_declaration(self):
        stmt = DeclarationStatement(
            var_type="int",
            variable=VariableExpression("i"),
            initialization=ConstantExpression(1)
        )
        code = CCodeGenerator().generate(stmt)
        assert code == "int i = 1;"

    def test_assignment(self):
        stmt = AssignmentStatement(
            left_term=VariableExpression("i"),
            right_term=VariableExpression("j")
        )
        code = CCodeGenerator().generate(stmt)
        assert code == "i = j;"

    def test_simple_for_loop(self):
        stmt = ForLoopStatement(
            body=[
                DeclarationStatement(
                    var_type="int",
                    variable=VariableExpression("x"),
                    initialization=ConstantExpression(1)
                ),
                AssignmentStatement(
                    left_term=VariableExpression(
                        "a", [VariableExpression("i")]),
                    right_term=Expression([
                        VariableExpression("x"),
                        ConstantExpression(1)
                    ], ["+"])
                )
            ],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        code = CCodeGenerator().generate(stmt)
        assert code == "for(int i = 1; i <= N; i++) {\n    int x = 1;\n    a[i] = x + 1;\n}"

    def test_nested_for_loop(self):
        inner_loop = ForLoopStatement(
            body=[
                DeclarationStatement(
                    var_type="int",
                    variable=VariableExpression("x"),
                    initialization=ConstantExpression(1)
                ),
                AssignmentStatement(
                    left_term=VariableExpression(
                        "a", [VariableExpression("j")]),
                    right_term=Expression([
                        VariableExpression("x"),
                        ConstantExpression(1)
                    ], ["+"])
                )
            ],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("M"),
            is_parallel=True
        )
        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        code = CCodeGenerator().generate(outer_loop)
        assert code == "for(int i = 1; i <= N; i++) {\n    #pragma omp parallel for\n    for(int j = 1; j <= M; j++) {\n        int x = 1;\n        a[j] = x + 1;\n    }\n}"



from __future__ import annotations

from abc import ABC


class Expression():

    def __init__(self,
                 terms: tuple[Expression],
                 operators: tuple[str]
                 ):
        if not len(terms) > 0:
            raise ValueError(
                "There must be at least one term in the expression!")
        self._terms = terms
        if not len(terms) - 1 == len(operators):
            raise ValueError(
                "Number of operators must be the number of terms minus one!")
        self._operators = operators

    @property
    def terms(self) -> tuple[Expression]:
        return self._terms

    @property
    def operators(self) -> tuple[str]:
        return self._operators

    def is_single(self) -> bool:
        return len(self.terms) == 1

    def is_constant(self) -> bool:
        return isinstance(self, ConstantExpression)

    def is_variable(self) -> bool:
        return isinstance(self, VariableExpression)

    def stringify(self) -> str:
        res = str(self.terms[0])
        for i in range(len(self.operators)):
            res += f" {self.operators[i]} {self.terms[i+1]}"
        return res

    def __str__(self):
        return self.stringify()

    def __repr__(self):
        return self.stringify()


class GroupingExpression(Expression):

    def stringify(self) -> str:
        return f"({super().stringify()})"


class SingleExpression(Expression, ABC):

    def __init__(self):
        super().__init__((self,), ())


class ConstantExpression(SingleExpression):

    def __init__(self, value: int or float):
        super().__init__()
        self._value = value

    @property
    def value(self) -> int or float:
        return self._value

    def stringify(self) -> str:
        return str(self.value)


class VariableExpression(SingleExpression):

    # TODO fix with Optional
    def __init__(self, name: str, indexes: tuple[Expression] = ()):
        super().__init__()
        self._name = name
        self._indexes = indexes

    @property
    def name(self) -> str:
        return self._name

    @property
    def indexes(self) -> tuple[Expression]:
        return self._indexes

    def is_simple(self) -> bool:
        return len(self.indexes) == 0

    def stringify(self) -> str:
        return f"{self.name}{''.join([f'[{i}]' for i in self.indexes])}"


class FunctionExpression(SingleExpression):

    def __init__(self, name: str, args: tuple[Expression] = ()):
        super().__init__()
        self._name = name
        self._args = args

    @property
    def name(self) -> str:
        return self._name

    @property
    def args(self) -> tuple[Expression]:
        return self._args

    def stringify(self) -> str:
        return f"{self.name}({', '.join([str(t) for t in self.args])})"


class UnaryExpression(Expression):

    def __init__(self, term: Expression, unary_operator: str):
        super().__init__([term], [])
        self._unary_operator = unary_operator

    @property
    def unary_operator(self) -> str:
        return self._unary_operator

    def stringify(self) -> str:
        return f"{self.unary_operator}{str(self.terms[0])}"


def extract_variable_expressions(expr: Expression) -> tuple[VariableExpression]:
    variable_expressions = []
    for subexpr in expr.terms:
        if isinstance(subexpr, VariableExpression):
            variable_expressions.append(subexpr)
        if isinstance(subexpr, FunctionExpression):
            for arg in subexpr.args:
                variable_expressions.extend(extract_variable_expressions(arg))
        elif not isinstance(subexpr, SingleExpression):
            variable_expressions.extend(extract_variable_expressions(subexpr))
    return tuple(variable_expressions)


def divide_variable_expressions_by_name(
    expressions: tuple[VariableExpression]
) -> dict[str, list[Expression]]:
    expressions_dict = {}
    for expr in expressions:
        exprs = expressions_dict.setdefault(expr.name, [])
        exprs.append(expr)
    return expressions_dict



import pytest

from opoly.variables import (
    SimpleVariableOccurrence,
    VariableOccurrenceType
)


class TestSimpleVariableOccurrence():

    def test_simple(self):
        occ = SimpleVariableOccurrence(
            name="a",
            vtype=VariableOccurrenceType.GENERATION,
            indexes=(
                ("i", 0),
                ("j", -1)
            )
        )
        assert str(occ) == "[GEN] STM a[i+0][j-1]"

    def test_indexes_names(self):
        occ = SimpleVariableOccurrence(
            name="a",
            vtype=VariableOccurrenceType.GENERATION,
            indexes=(
                ("i", 1),
                ("j", -1)
            )
        )
        assert occ.indexes_names == ("i", "j")

    def test_offsets(self):
        occ = SimpleVariableOccurrence(
            name="a",
            vtype=VariableOccurrenceType.GENERATION,
            indexes=(
                ("i", 1),
                ("j", -1)
            )
        )
        assert occ.offsets == (1, -1)



import pytest

from opoly.expressions import (
    Expression,
    SingleExpression,
    GroupingExpression,
    ConstantExpression,
    VariableExpression
)

from opoly.statements import (
    StatementType,
    AssignmentStatement,
    DeclarationStatement,
    ForLoopStatement,
    divide_assignments,
    prune_expressions
)


class TestAssignmentStatement():

    def test_simple_assignment(self):
        stmt = AssignmentStatement(
            left_term=VariableExpression("a"),
            right_term=ConstantExpression(1)
        )
        assert stmt.stype == StatementType.ASSIGNMENT
        assert str(stmt) == "a = 1"

    def test_complex_assigment(self):
        stmt = AssignmentStatement(
            left_term=VariableExpression("a",
                                         [VariableExpression("i"), VariableExpression("j")]),
            right_term=Expression([
                VariableExpression("a",
                                   [Expression([VariableExpression("i"), ConstantExpression(1)], ["+"]),
                                    Expression([VariableExpression("j"), ConstantExpression(1)], ["+"])]),
                VariableExpression("a",
                                   [Expression([VariableExpression("i"), ConstantExpression(1)], ["-"]),
                                    Expression([VariableExpression("j"), ConstantExpression(1)], ["-"])])
            ], ["+"])
        )
        assert str(stmt) == "a[i][j] = a[i + 1][j + 1] + a[i - 1][j - 1]"

    def test_simple_divide_assignment(self):
        stmt = AssignmentStatement(
            left_term=VariableExpression("a",
                                         [VariableExpression("i"), VariableExpression("j")]),
            right_term=Expression([
                VariableExpression("a",
                                   [Expression([VariableExpression("i"), ConstantExpression(1)], ["+"]),
                                    Expression([VariableExpression("j"), ConstantExpression(1)], ["+"])]),
                VariableExpression("a",
                                   [Expression([VariableExpression("i"), ConstantExpression(1)], ["-"]),
                                    Expression([VariableExpression("j"), ConstantExpression(1)], ["-"])])
            ], ["+"])
        )
        gens, uses = divide_assignments([stmt])
        assert list(str(g) for g in gens) == ["a[i][j]"]
        assert list(str(u) for u in uses) == [
            "a[i + 1][j + 1]", "a[i - 1][j - 1]"]
        gens, uses = prune_expressions(gens, uses)
        assert list(str(g) for g in gens) == ["a[i][j]"]
        assert list(str(u) for u in uses) == [
            "a[i + 1][j + 1]", "a[i - 1][j - 1]"]

    def test_complex_divide_and_prune_assignments(self):
        stmt1 = AssignmentStatement(
            left_term=VariableExpression("a",
                                         [VariableExpression("i"), VariableExpression("j")]),
            right_term=Expression([
                GroupingExpression([
                    VariableExpression("b",
                                       [Expression([VariableExpression("i"), ConstantExpression(1)], ["+"]),
                                        Expression([VariableExpression("j"), ConstantExpression(1)], ["+"])]),
                    VariableExpression("b",
                                       [Expression([VariableExpression("i"), ConstantExpression(1)], ["-"]),
                                        Expression([VariableExpression("j"), ConstantExpression(1)], ["-"])])
                ], ["+"]),
                ConstantExpression(2.0)
            ], ["/"])
        )
        assert str(
            stmt1) == "a[i][j] = (b[i + 1][j + 1] + b[i - 1][j - 1]) / 2.0"

        stmt2 = AssignmentStatement(
            left_term=VariableExpression("b",
                                         [VariableExpression("i"), VariableExpression("j")]),
            right_term=Expression([
                GroupingExpression([
                    VariableExpression("c",
                                       [Expression([VariableExpression("i"), ConstantExpression(1)], ["+"]),
                                        Expression([VariableExpression("j"), ConstantExpression(1)], ["+"])]),
                    VariableExpression("c",
                                       [Expression([VariableExpression("i"), ConstantExpression(1)], ["-"]),
                                        Expression([VariableExpression("j"), ConstantExpression(1)], ["-"])])
                ], ["+"]),
                ConstantExpression(2.0)
            ], ["/"])
        )
        assert str(
            stmt2) == "b[i][j] = (c[i + 1][j + 1] + c[i - 1][j - 1]) / 2.0"

        gens, uses = divide_assignments([stmt1, stmt2])
        assert list(str(g) for g in gens) == [
            "a[i][j]", "b[i][j]"
        ]
        assert list(str(u) for u in uses) == [
            "b[i + 1][j + 1]", "b[i - 1][j - 1]",
            "c[i + 1][j + 1]", "c[i - 1][j - 1]"
        ]
        pruned_gens, pruned_uses = prune_expressions(gens, uses)
        assert list(str(g) for g in pruned_gens) == ["b[i][j]"]
        assert list(str(u) for u in pruned_uses) == [
            "b[i + 1][j + 1]", "b[i - 1][j - 1]"
        ]

        stmt3 = AssignmentStatement(
            left_term=VariableExpression("a",
                                         [Expression([VariableExpression("i"), ConstantExpression(1)], ["+"]),
                                          VariableExpression("j")]),
            right_term=Expression([
                GroupingExpression([
                    VariableExpression("c",
                                       [Expression([VariableExpression("i"), ConstantExpression(1)], ["+"]),
                                        Expression([VariableExpression("j"), ConstantExpression(1)], ["+"])]),
                    VariableExpression("c",
                                       [Expression([VariableExpression("i"), ConstantExpression(1)], ["-"]),
                                        Expression([VariableExpression("j"), ConstantExpression(1)], ["-"])])
                ], ["+"]),
                ConstantExpression(2.0)
            ], ["/"])
        )
        assert str(
            stmt3) == "a[i + 1][j] = (c[i + 1][j + 1] + c[i - 1][j - 1]) / 2.0"

        gens, uses = divide_assignments([stmt1, stmt2, stmt3])
        assert list(str(g) for g in gens) == [
            "a[i][j]", "b[i][j]", "a[i + 1][j]"
        ]
        assert list(str(u) for u in uses) == [
            "b[i + 1][j + 1]", "b[i - 1][j - 1]",
            "c[i + 1][j + 1]", "c[i - 1][j - 1]",
            "c[i + 1][j + 1]", "c[i - 1][j - 1]"
        ]
        pruned_gens, pruned_uses = prune_expressions(gens, uses)
        assert list(str(g) for g in pruned_gens) == [
            "a[i][j]", "b[i][j]", "a[i + 1][j]"
        ]
        assert list(str(u) for u in pruned_uses) == [
            "b[i + 1][j + 1]", "b[i - 1][j - 1]"
        ]


class TestDeclarationStatement():
    
    def test_simple_declaration(self):
        stmt = DeclarationStatement(
            var_type="int",
            variable=VariableExpression("i")
        )
        assert stmt.stype == StatementType.DECLARATION
        assert str(stmt) == "int i"
    
    def test_complex_declaration(self):
        stmt = DeclarationStatement(
            var_type="int",
            variable=VariableExpression("i"),
            initialization=Expression([
                VariableExpression("j"),
                ConstantExpression(2)
            ], ["-"])
        )
        assert stmt.stype == StatementType.DECLARATION
        assert str(stmt) == "int i = j - 2"

class TestForLoopStatement():

    def test_plain_loop(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        assert str(loop) == "FOR i = 1...N\na[i] = b[i]"

    def test_nested_loop(self):
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("j")]),
                                      VariableExpression("b", [VariableExpression("j")]))],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("M")
        )

        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        assert str(outer_loop) == "FOR i = 1...N\nFOR j = 1...M\na[j] = b[j]"

    def test_empty_loop(self):
        with pytest.raises(ValueError):
            _ = ForLoopStatement(body=[],
                                 index=VariableExpression("i"),
                                 lowerbound=ConstantExpression(1),
                                 upperbound=ConstantExpression(10)
                                 )



import numpy as np

from opoly.modules.parser import PseudocodeForLoopParser
from opoly.modules.checker import LamportForLoopChecker
from opoly.modules.detector import LamportLoopDependenciesDetector
from opoly.modules.scheduler import LamportCPScheduler
from opoly.modules.allocator import LamportCPAllocator
from opoly.modules.scanner import FourierMotzkinScanner


class TestPseudocodeForLoopParserToFourierMotzkinScanner():

    def _pipeline_parser_scheduler(self, code: str) -> np.ndarray:
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        _, _ = LamportForLoopChecker().check(loop)
        deps = LamportLoopDependenciesDetector().extract_dependencies(loop)
        deps_np = np.array(list(list(d.converted_values) for d in deps))
        schedule, _ = LamportCPScheduler().schedule(deps_np)
        allocation, _ = LamportCPAllocator().allocate(schedule)
        reindexed_loop = FourierMotzkinScanner().reindex(loop, allocation)
        return reindexed_loop

    def test_1d_loop(self):
        code = "FOR i FROM 0 TO N { STM a[i]=a[i+1]; }"
        reindexed_loop = self._pipeline_parser_scheduler(code)
        assert reindexed_loop is not None

    def test_2d_loop_example1(self):
        code = "FOR i FROM 1 TO N-1 { FOR j FROM 1 TO M-1 { STM a[i][j] = (a[i-1][j] + a[i][j] + a[i][j-1]) / 3.0; } }"
        reindexed_loop = self._pipeline_parser_scheduler(code)
        assert reindexed_loop is not None

    def test_2d_loop_example2(self):
        code = "FOR i FROM 1 TO N-1 { FOR j FROM 1 TO M-1 { STM a[j] = (a[j-1] + a[j] + a[j+1]) / 3.0; } }"
        reindexed_loop = self._pipeline_parser_scheduler(code)
        assert reindexed_loop is not None

    def test_3d_loop_example5(self):
        code = """
        FOR i FROM 1 TO N-1 {
            FOR j FROM 1 TO M-2 {
                FOR k FROM 1 TO L-2 {
                    STM u[j][k] = (u[j+1][k] + u[j][k+1] + u[j-1][k] + u[j][k-1]) * 0.25;
                }
            }
        }
        """
        reindexed_loop = self._pipeline_parser_scheduler(code)
        assert reindexed_loop is not None



from opoly.expressions import ConstantExpression, VariableExpression
from opoly.modules.parser import (
    parse_constant_expression,
    parse_variable_expression,
    parse_grouping_expression,
    parse_simple_function_expression,
    parse_function_expression,
    parse_operator,
    parse_expression,
    parse_assignment_statement,
    parse_unary_expression,
    PseudocodeForLoopParser
)


class TestParsingFunctions():

    def test_constant_parser(self):
        goods = {
            "0": (0, ""),
            "1": (1, ""),
            "0.5": (0.5, ""),
            "3.0": (3.0, ""),
            "10": (10, ""),
            "1000000": (1000000, ""),
            "1 / 2": (1, " / 2"),
            "1a": (1, "a")
        }
        bads = [
            "-1",
            "a",
            "",
            "/"
        ]
        for code, res in goods.items():
            parsed, rem = parse_constant_expression(code)
            assert parsed is not None
            assert parsed.value == ConstantExpression(res[0]).value
            assert type(parsed.value) == type(ConstantExpression(res[0]).value)
            assert rem == res[1]
        for code in bads:
            parsed, rem = parse_constant_expression(code)
            assert parsed is None

    def test_simple_variable_parser(self):
        goods = {
            "a": ("a", ""),
            "alpha": ("alpha", ""),
            "a_b": ("a_b", ""),
            "a + 1": ("a", " + 1"),
            "a+1": ("a", "+1"),
            "a1": ("a", "1")
        }
        bads = {
            "": "Expected variable expression",
            "0": "Expected variable expression",
            "1": "Expected variable expression",
            "-1": "Expected variable expression",
            "1a": "Expected variable expression",
            "[asd]": "Expected variable expression"
        }
        for code, res in goods.items():
            parsed, rem = parse_variable_expression(code)
            assert parsed is not None
            assert str(parsed) == str(VariableExpression(res[0]))
            assert rem == res[1]
        for code, err in bads.items():
            parsed, error = parse_variable_expression(code)
            assert parsed is None
            assert error == err

    def test_indexed_variable_parser(self):
        goods = {
            "a[1]": ("a[1]", ""),
            "a[1][1]": ("a[1][1]", ""),
            "a[i]": ("a[i]", ""),
            "a[i+1]": ("a[i + 1]", ""),
            "a[i+1][j+1] + 1": ("a[i + 1][j + 1]", " + 1"),
            "a[a[i]]": ("a[a[i]]", ""),
            "a[i] + 1": ("a[i]", " + 1"),
            "a[i]+1": ("a[i]", "+1"),
            "a]": ("a", "]")
        }
        bads = {
            "a[": "Expected expression",
            "a[]": "Unsupported expression term",
            "a[[]": "Unsupported expression term",
            "a[a[i]": "Expected ]"
        }
        for code, res in goods.items():
            parsed, rem = parse_variable_expression(code)
            assert parsed is not None
            assert str(parsed) == str(VariableExpression(res[0]))
            assert rem == res[1]
        for code, err in bads.items():
            parsed, error = parse_variable_expression(code)
            assert parsed is None
            assert error == err

    def test_operator_parser(self):
        goods = {
            "+": ("+", ""),
            "-": ("-", ""),
            "*": ("*", ""),
            "/": ("/", ""),
            "/ i": ("/", " i"),
        }
        bads = {
            "": "Expected operator"
        }
        for code, res in goods.items():
            parsed, rem = parse_operator(code)
            assert parsed is not None
            assert str(parsed) == res[0]
            assert rem == res[1]
        for code, err in bads.items():
            parsed, error = parse_operator(code)
            assert parsed is None
            assert error == err

    def test_grouping_parser(self):
        goods = {
            "(a)": ("(a)", ""),
            "(a + b)": ("(a + b)", ""),
            "(a +b)": ("(a + b)", ""),
            "(a+b)": ("(a + b)", ""),
            "(a[i] + 1)": ("(a[i] + 1)", ""),
            "(a[i] + b[i][j+1]) / 3.0": ("(a[i] + b[i][j + 1])", " / 3.0")
        }
        bads = {
            "": "Expected (",
            "(": "Expected expression",
            ")": "Expected (",
            "(a": "Expected )",
            "(a a)": "Unsupported operator",
            "(a + )": "Unsupported expression term"
        }
        for code, res in goods.items():
            parsed, rem = parse_grouping_expression(code)
            assert parsed is not None
            assert str(parsed) == res[0]
            assert rem == res[1]
        for code, err in bads.items():
            parsed, error = parse_grouping_expression(code)
            assert parsed is None
            assert error == err

    def test_simple_function_parser(self):
        goods = {
            "max(a)": ("max(a)", ""),
            "max(a) + i": ("max(a)", " + i"),
            "max(a, b, c) + i": ("max(a, b, c)", " + i"),
        }
        bads = {
            "": "Expected function expression",
            "max": "Expected function expression",
            "max(a": "Expected function expression",
            "max(,a)": "Function expression not well formatted",
            "max(a b)": "Unsupported operator",
        }
        for code, res in goods.items():
            parsed, rem = parse_simple_function_expression(code)
            assert parsed is not None
            assert str(parsed) == res[0]
            assert rem == res[1]
        for code, err in bads.items():
            parsed, error = parse_simple_function_expression(code)
            assert parsed is None
            assert error == err
    
    def test_function_parser(self):
        goods = {
            "max(a)": ("max(a)", ""),
            "max(a) + i": ("max(a)", " + i"),
            "max(a, b, c) + i": ("max(a, b, c)", " + i"),
            "ceil(max(a, b), min(a, b, c))": ("ceil(max(a, b), min(a, b, c))", ""),
            "ceil(max(a, b/2), min(a, b, c))": ("ceil(max(a, b / 2), min(a, b, c))", "")
        }
        bads = {
            "": "Expected function expression",
            "max": "Expected (",
            "max(a": "Expected ,",
            "max(,a)": "Unsupported expression term",
            "max(a b)": "Unsupported operator",
        }
        for code, res in goods.items():
            parsed, rem = parse_function_expression(code)
            assert parsed is not None
            assert str(parsed) == res[0]
            assert rem == res[1]
        for code, err in bads.items():
            parsed, error = parse_function_expression(code)
            assert parsed is None
            assert error == err
    
    def test_unary_expression_parser(self):
        goods = {
            "-1": ("-1", ""),
            "-n": ("-n", ""),
            "-(n+1)": ("-(n + 1)", ""),
            "-(a[1]-1)": ("-(a[1] - 1)", ""),
            "- a": ("-a", "")
        }
        bads = {
            "--a": "Unsupported unary operator",
            "- -a": "Unsupported unary operator"
        }
        for code, res in goods.items():
            parsed, rem = parse_unary_expression(code)
            assert parsed is not None
            assert str(parsed) == res[0]
            assert rem == res[1]
        for code, err in bads.items():
            parsed, error = parse_unary_expression(code)
            assert parsed is None
            assert error == err

    def test_expression_parser(self):
        goods = {
            "1": ("1", ""),
            "a + i": ("a + i", ""),
            "a[i] + b[i]": ("a[i] + b[i]", ""),
            "a[i] + b[i] / 3": ("a[i] + b[i] / 3", ""),
            "a[i+1][j+1] + a[i-1][j-1] / 3": ("a[i + 1][j + 1] + a[i - 1][j - 1] / 3", ""),
            "(a[i+1][j+1] + a[i-1][j-1]) / 3.0": ("(a[i + 1][j + 1] + a[i - 1][j - 1]) / 3.0", ""),
            "(-n + i - 1) / 2.0": ("(-n + i - 1) / 2.0", ""),
            "floor(min(1, (1.0/3.0) * (2 * a + b), 2 * i), a + c)": ("floor(min(1, (1.0 / 3.0) * (2 * a + b), 2 * i), a + c)", "")
        }
        bads = {
            "": "Expected expression",
            "a +": "Expected expression",
            "a+": "Expected expression",
            "a+[": "Unsupported expression term",
            "[": "Unsupported expression term",
            "+": "Unsupported expression term",
            "a a": "Unsupported operator",
            "]": "Unsupported expression term",
            "a]": "Unsupported operator",
            "[]": "Unsupported expression term",
            "(": "Expected expression"
        }
        for code, res in goods.items():
            parsed, rem = parse_expression(code)
            assert parsed is not None
            assert str(parsed) == res[0]
            assert rem == res[1]
        for code, err in bads.items():
            parsed, error = parse_expression(code)
            assert parsed is None
            assert error == err

        parsed, err = parse_expression("a,", term_char=")")
        assert parsed is None
        assert err == "Expected )"

    def test_assignment_statement_parser(self):
        goods = {
            "a = a + i": ("a = a + i", None),
            "a[i] = a + i": ("a[i] = a + i", None),
            "a[i+1] = a + i": ("a[i + 1] = a + i", None),
            "a = a[i] + b[i]": ("a = a[i] + b[i]", None),
            "a =a[i] + b[i] / 3": ("a = a[i] + b[i] / 3", None),
            "a=a[i+1][j+1] + a[i-1][j-1] / 3": ("a = a[i + 1][j + 1] + a[i - 1][j - 1] / 3", None),
            "a= (a[i+1][j+1] + a[i-1][j-1]) / 3.0": ("a = (a[i + 1][j + 1] + a[i - 1][j - 1]) / 3.0", None),
        }
        bads = {
            "": "Expected = operator",
            "a = b = c": "Expected only one = operator in assignment",
            "= a": "Expected variable expression",
            "a = ": "Expected expression",
            "a + a = a": "Expected only one variable expression for left term",
            "a = bc[": "Expected expression",
            "a = asd + asd +": "Expected expression",
            "a = ]": "Unsupported expression term",
        }
        for code, res in goods.items():
            parsed, rem = parse_assignment_statement(code)
            print(rem)
            assert parsed is not None
            assert str(parsed) == res[0]
            assert rem == res[1]
        for code, err in bads.items():
            parsed, error = parse_assignment_statement(code)
            assert parsed is None
            assert error == err


class TestPseudocodeForLoopParser():

    def test_simple_parse_loop_body(self):
        code = "STM a[i] = a[i+1];"
        stmts, err = PseudocodeForLoopParser().parse_loop_body(code)
        print(err)
        assert stmts is not None
        assert len(stmts) == 1
        assert str(stmts[0]) == "a[i] = a[i + 1]"

    def test_long_parse_loop_body(self):
        code = "STM a = 10;\nSTM b = a;\nSTM c[x] = a+b;"
        stmts, err = PseudocodeForLoopParser().parse_loop_body(code)
        print(err)
        assert stmts is not None
        assert len(stmts) == 3
        assert str(stmts[0]) == "a = 10"
        assert str(stmts[1]) == "b = a"
        assert str(stmts[2]) == "c[x] = a + b"

    def test_parse_1d_loop(self):
        code = "FOR i FROM 0 TO n { STM a[i]=a[i+1]; }"
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        print(err)
        assert loop is not None
        assert len(loop.body) == 1
        assert str(loop.index) == "i"
        assert str(loop.lowerbound) == "0"
        assert str(loop.upperbound) == "n"
        assert str(loop.step) == "1"
        assert str(loop.body[0]) == "a[i] = a[i + 1]"

    def test_parse_plain_nested_2d_loop(self):
        code = "FOR i FROM 1 TO N - 1 STEP 1 { FOR j FROM 2 TO M-1 { STM a[j] = (a[j-1] + a[j] + a[j+1]) / 3.0; } }"
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        print(err)
        assert loop is not None
        assert len(loop.body) == 1
        assert str(loop.index) == "i"
        assert str(loop.lowerbound) == "1"
        assert str(loop.upperbound) == "N - 1"
        assert str(loop.step) == "1"
        inner_loop = loop.body[0]
        assert inner_loop is not None
        assert len(inner_loop.body) == 1
        assert str(inner_loop.index) == "j"
        assert str(inner_loop.lowerbound) == "2"
        assert str(inner_loop.upperbound) == "M - 1"
        assert str(inner_loop.step) == "1"
        assert str(
            inner_loop.body[0]) == "a[j] = (a[j - 1] + a[j] + a[j + 1]) / 3.0"

    def test_wrong_inner_loop(self):
        code = "FOR i FROM 1 TO N - 1 STEP 1 { FOR j FROM 2 TO M { STM a[j] == (a[j-1] + a[j] + a[j+1]) / 3.0; } }"
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        assert loop is None

    def test_unsupported_inner_statement_loop(self):
        code = "FOR i FROM 0 TO n { asd }"
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        assert loop is None

    def test_expected_loop(self):
        code = ""
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        assert loop is None
        code = "asdasd"
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        assert loop is None

    def test_wrong_loop_index(self):
        code = "FOR 2 FROM 1 TO N {a = 1;}"
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        assert loop is None

    def test_wrong_lowerbound(self):
        code = "FOR i FROM 2**2 TO N {STM a = 1;}"
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        assert loop is None

    def test_wrong_upperbound(self):
        code = "FOR i FROM 1 TO N//2 {STM a = 1;}"
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        assert loop is None

    def test_wrong_step(self):
        code = "FOR i FROM 1 TO N/2 STEP 2**2 {STM a = 1;}"
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        assert loop is None

    def test_empty_body(self):
        code = "FOR i FROM 1 TO N/2 STEP j { }"
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        assert loop is None
    
    def test_complex_loop(self):
        code = """
        FOR i FROM 0 TO N {
            STM a[i][0] = 0;
            FOR j FROM 1 TO M-1 {
                STM a[i][j] = a[i][j-1] + a[i][j+1];
                STM a[j][j] = 0;
            }
            STM a[N][M] = a[0][0];
        }
        """
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        assert loop is not None



from opoly.statements import ForLoopStatement, AssignmentStatement, DeclarationStatement
from opoly.expressions import VariableExpression, ConstantExpression, Expression, GroupingExpression

from opoly.modules.checker import is_perfectly_nested_loop, is_plain_loop, is_recursively_plain_loop, LamportForLoopChecker


class TestForLoopStatementChecks():

    def test_plain_loop(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        assert is_plain_loop(loop)
        assert LamportForLoopChecker().check(loop)[0]

    def test_nested_loop(self):
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("j")]),
                                      VariableExpression("b", [VariableExpression("j")]))],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("M")
        )

        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        assert is_plain_loop(outer_loop)
        assert is_perfectly_nested_loop(outer_loop)
        assert LamportForLoopChecker().check(outer_loop)[0]

    def test_not_simple_index(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("a", [VariableExpression("i")]),
            lowerbound=VariableExpression("i"),
            upperbound=VariableExpression("M")
        )
        assert not is_plain_loop(loop)
        assert not LamportForLoopChecker().check(loop)[0]

    def test_not_constant_lowerbound(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=VariableExpression("j"),
            upperbound=VariableExpression("N")
        )
        assert not is_plain_loop(loop)
        assert not LamportForLoopChecker().check(loop)[0]

    def test_not_simple_upperbound(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("a", [VariableExpression("i")])
        )
        assert not is_plain_loop(loop)
        assert not LamportForLoopChecker().check(loop)[0]

    def test_not_constant_step(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N"),
            step=VariableExpression("j")
        )
        assert not is_plain_loop(loop)
        assert not LamportForLoopChecker().check(loop)[0]

    def test_not_singular_increment_step(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N"),
            step=ConstantExpression(2)
        )
        assert not is_plain_loop(loop)
        assert not LamportForLoopChecker().check(loop)[0]

    def test_perfectly_nested_loop(self):
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("j")]),
                                      VariableExpression("b", [VariableExpression("j")]))],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("M")
        )

        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )

        assert is_perfectly_nested_loop(outer_loop)
        assert LamportForLoopChecker().check(outer_loop)[0]

    def test_not_perfectly_nested_loop(self):
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("j")]),
                                      VariableExpression("b", [VariableExpression("j")]))],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("M")
        )

        outer_loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")])),
                  inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )

        assert not is_perfectly_nested_loop(outer_loop)
        assert not LamportForLoopChecker().check(outer_loop)[0]

    def test_plain_nested_loop(self):
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("j")]),
                                      VariableExpression("b", [VariableExpression("j")]))],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("M")
        )

        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )

        assert is_plain_loop(outer_loop)
        assert is_perfectly_nested_loop(outer_loop)
        assert LamportForLoopChecker().check(outer_loop)[0]

    def test_not_plain_nested_loop(self):
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("j")]),
                                      VariableExpression("b", [VariableExpression("j")]))],
            index=VariableExpression("j"),
            lowerbound=VariableExpression("i"),
            upperbound=VariableExpression("M")
        )

        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )

        assert not is_recursively_plain_loop(outer_loop)
        assert is_perfectly_nested_loop(outer_loop)


class TestLamportForLoopChecker():

    def test_simple_2d_loop(self):
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(
                left_term=VariableExpression("a", [VariableExpression("j")]),
                right_term=Expression([
                    GroupingExpression([
                        VariableExpression("a", [Expression([
                            VariableExpression("j"), ConstantExpression(1)], "-")]),
                        VariableExpression("a", [VariableExpression("j")]),
                        VariableExpression("a", [Expression([
                            VariableExpression("j"), ConstantExpression(1)], "+")])
                    ], ["+", "+"]),
                    ConstantExpression(3.0)
                ], "/")
            )],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression(
                [VariableExpression("M"), ConstantExpression(1)], ["-"])
        )

        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )

        assert is_plain_loop(outer_loop)
        assert is_perfectly_nested_loop(outer_loop)
        assert LamportForLoopChecker().check(outer_loop)[0]

    def test_simple_3d_loop(self):
        ajk = VariableExpression("a", [
            VariableExpression("j"),
            VariableExpression("k")
        ])
        ajminus1k = VariableExpression("a", [
            Expression([VariableExpression("j"),
                        ConstantExpression(1)], ["-"]),
            VariableExpression("k")
        ])
        ajminus1kminus1 = VariableExpression("a", [
            Expression([VariableExpression("j"),
                        ConstantExpression(1)], ["-"]),
            Expression([VariableExpression("k"),
                        ConstantExpression(1)], ["-"]),
        ])

        right = Expression([
            ajk, ajminus1k, ajminus1kminus1
        ], ["+", "-"])
        left = ajk

        kloop = ForLoopStatement(
            body=[AssignmentStatement(left, right)],
            index=VariableExpression("k"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("l")
        )

        jloop = ForLoopStatement(
            body=[kloop],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("m")
        )

        iloop = ForLoopStatement(
            body=[jloop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("n")
        )

        assert is_plain_loop(iloop)
        assert is_perfectly_nested_loop(iloop)
        assert LamportForLoopChecker().check(iloop)[0]

    def test_not_simple_indexes(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i", [ConstantExpression(0)]),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(loop)
        assert not res
        assert error == "Loop indexes are not simple!"

    def test_not_all_distinct_indexes(self):
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("M")
        )
        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(outer_loop)
        assert not res
        assert error == "Not all index names in loop are distinct!"

    def test_not_all_assigments(self):
        loop = ForLoopStatement(
            body=[DeclarationStatement(var_type="int", variable=VariableExpression("x")),
                  AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(loop)
        assert not res
        assert error == "Not all statements in loop body are assignments!"

    def test_not_all_left_variables(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(ConstantExpression(1),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(loop)
        assert not res
        assert error == "Not all left sides of statements in loop body are variable expressions!"

    def test_not_all_non_simple_generations(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a"),
                                      VariableExpression("a", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(loop)
        assert not res
        assert error == "All variable generations must be non-simple!"

    def test_not_all_same_name_simple_or_indexed(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      Expression([
                                          VariableExpression(
                                              "a", [VariableExpression("i")]),
                                          VariableExpression("a")
                                      ], ["+"]
            ))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(loop)
        assert not res
        assert error == "Same name variables must be all simple or all non-simple!"

    def test_variable_index_not_present(self):
        wrong = VariableExpression("a", [VariableExpression("j")])
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      wrong)],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(loop)
        assert not res
        assert error == f"Variable expression ({str(wrong)}) has an index not present in loop indexes!"

    def test_not_unique_variable_indexes(self):
        wrong = VariableExpression(
            "a", [VariableExpression("j"), VariableExpression("j")])
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(
                left_term=wrong,
                right_term=Expression([
                    GroupingExpression([
                        VariableExpression("a", [Expression([
                            VariableExpression("j"), ConstantExpression(1)], "-"), VariableExpression("j")]),
                        VariableExpression(
                            "a", [VariableExpression("j"), VariableExpression("j")]),
                        VariableExpression("a", [Expression([
                            VariableExpression("j"), ConstantExpression(1)], "+"), VariableExpression("j")])
                    ], ["+", "+"]),
                    ConstantExpression(3.0)
                ], "/")
            )],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression(
                [VariableExpression("M"), ConstantExpression(1)], ["-"])
        )
        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(outer_loop)
        assert not res
        assert error == f"Variable expression ({str(wrong)}) indexes are not unique!"

    def test_wrong_missing_index(self):
        wrong = VariableExpression("a", [VariableExpression("i")])
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(
                left_term=VariableExpression("a", [VariableExpression("i")]),
                right_term=Expression([
                    GroupingExpression([
                        VariableExpression("a", [Expression([
                            VariableExpression("i"), ConstantExpression(1)], "-")]),
                        VariableExpression("a", [VariableExpression("i")]),
                        VariableExpression("a", [Expression([
                            VariableExpression("i"), ConstantExpression(1)], "+")])
                    ], ["+", "+"]),
                    ConstantExpression(3.0)
                ], "/")
            )],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression(
                [VariableExpression("M"), ConstantExpression(1)], ["-"])
        )
        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(outer_loop)
        assert not res
        assert error == f"Variable expression ({str(wrong)}) has a missing index which is not ({outer_loop.index})!"

    def test_not_integer_constant(self):
        wrong = VariableExpression(
            "a", [Expression([VariableExpression("i"), ConstantExpression(1.5)], ["+"])])
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      wrong)],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(loop)
        assert not res
        assert error == f"Variable expression ({str(wrong)}) has a non-integer constant in one of the indexes!"

    def test_not_simple_variable_indexes(self):
        wrong = VariableExpression(
            "a", [VariableExpression("i", [ConstantExpression(1)])])
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      wrong)],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(loop)
        assert not res
        assert error == f"Variable expression ({str(wrong)}) has incorrect indexes!"

    def test_multiple_variable_indexes(self):
        wrong = VariableExpression(
            "a", [Expression([VariableExpression("i"), VariableExpression("j")], ["+"])])
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      wrong)],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(loop)
        assert not res
        assert error == f"Variable expression ({str(wrong)}) has incorrect indexes!"

    def test_inverted_order_index_expression(self):
        wrong = VariableExpression("a", [Expression([
            ConstantExpression(1), VariableExpression("j")], "-")])
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(
                left_term=VariableExpression("a", [VariableExpression("j")]),
                right_term=Expression([
                    GroupingExpression([
                        wrong,
                        VariableExpression("a", [VariableExpression("j")]),
                        VariableExpression("a", [Expression([
                            VariableExpression("j"), ConstantExpression(1)], "+")])
                    ], ["+", "+"]),
                    ConstantExpression(3.0)
                ], "/")
            )],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression(
                [VariableExpression("M"), ConstantExpression(1)], ["-"])
        )

        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(outer_loop)
        assert not res
        assert error == f"Variable expression ({str(wrong)}) has incorrect indexes!"

    def test_not_supported_index_operator(self):
        wrong = VariableExpression(
            "a", [Expression([VariableExpression("i"), ConstantExpression(2)], ["/"])])
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      wrong)],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        res, error = LamportForLoopChecker().check(loop)
        assert not res
        assert error == f"Variable expression ({str(wrong)}) has incorrect indexes!"

    def test_different_variable_indexes_order(self):
        ajk = VariableExpression("a", [
            VariableExpression("j"),
            VariableExpression("k")
        ])
        akminus1k = VariableExpression("a", [
            Expression([VariableExpression("k"),
                        ConstantExpression(1)], ["-"]),
            VariableExpression("j")
        ])
        ajminus1kminus1 = VariableExpression("a", [
            Expression([VariableExpression("j"),
                        ConstantExpression(1)], ["-"]),
            Expression([VariableExpression("k"),
                        ConstantExpression(1)], ["-"]),
        ])

        right = Expression([
            ajk, akminus1k, ajminus1kminus1
        ], ["+", "-"])
        left = ajk

        kloop = ForLoopStatement(
            body=[AssignmentStatement(left, right)],
            index=VariableExpression("k"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("l")
        )

        jloop = ForLoopStatement(
            body=[kloop],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("m")
        )

        iloop = ForLoopStatement(
            body=[jloop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("n")
        )

        res, error = LamportForLoopChecker().check(iloop)
        assert not res
        assert error == f"Variable expressions ({str(ajk)}) and ({str(akminus1k)}) have the same name but different indexes order!"



from abc import ABC, abstractmethod
import itertools

from opoly.statements import ForLoopStatement
from opoly.indexes import IndexDescriptor, IndexDescriptorType, IndexSet
from opoly.modules.checker import (
    extract_loop_indexes,
    get_inner_loop_statements,
    prune_expressions,
    divide_assignments,
    divide_variable_expressions_by_name,
    get_indexed_variable_simple_indexes,
    get_indexed_variable_index_constants_with_sign
)


class LoopDependenciesDetector():

    @abstractmethod
    def extract_dependencies(self, loop: ForLoopStatement) -> tuple[IndexSet]:
        pass


class LamportLoopDependenciesDetector(LoopDependenciesDetector):

    def generate_dependencies_from_couples(
        self,
        couples,
        indexes,
        indexes_pos_lookup,
        couples_index_pos_lookup
    ) -> list[IndexSet]:
        deps = []
        for gen1, gen2 in couples:
            gen1_consts = get_indexed_variable_index_constants_with_sign(gen1)
            gen2_consts = get_indexed_variable_index_constants_with_sign(gen2)
            dep_vector = tuple(c1.value - c2.value for c1,
                               c2 in zip(gen1_consts, gen2_consts))
            index_descriptors = [None] * len(indexes)
            for index in indexes:
                if index.name in couples_index_pos_lookup:  # pylint: disable=no-member
                    index_descriptors[indexes_pos_lookup[index.name]] = IndexDescriptor(  # pylint: disable=no-member
                        IndexDescriptorType.CONSTANT, value=dep_vector[
                            couples_index_pos_lookup[index.name]]  # pylint: disable=no-member
                    )
                else:
                    index_descriptors[indexes_pos_lookup[index.name]] = IndexDescriptor(  # pylint: disable=no-member
                        IndexDescriptorType.ANY
                    )
            deps.append(IndexSet(index_descriptors))
        return deps

    def extract_dependencies(self, loop: ForLoopStatement) -> tuple[IndexSet]:
        indexes = extract_loop_indexes(loop)
        inner_statements = get_inner_loop_statements(loop)
        lefts, rights = divide_assignments(inner_statements)
        generations, uses = prune_expressions(lefts, rights)
        uses = tuple(filter(lambda use: not use.is_simple(), uses))

        # Check variable indexes
        generations_by_name = divide_variable_expressions_by_name(generations)
        uses_by_name = divide_variable_expressions_by_name(uses)

        # Create index position lookup
        index_pos_lookup = {}
        for pos, index in enumerate(indexes):
            index_pos_lookup[index.name] = pos

        # Generate write-write dependencies
        write_write_index_sets = []
        for _, same_name_generations in generations_by_name.items():
            gen_indexes = get_indexed_variable_simple_indexes(
                same_name_generations[0])
            # Create gen index position lookup
            gen_index_pos_lookup = {}
            for pos, gen_index in enumerate(gen_indexes):
                gen_index_pos_lookup[gen_index.name] = pos
            # Create dependency index set between gen1 and gen2
            write_write_index_sets.extend(self.generate_dependencies_from_couples(
                itertools.product(same_name_generations,
                                  same_name_generations),
                indexes,
                index_pos_lookup,
                gen_index_pos_lookup
            ))

        # Generate write-read dependencies
        write_read_index_sets = []
        for var_name in generations_by_name.keys():
            gen_indexes = get_indexed_variable_simple_indexes(
                generations_by_name[var_name][0])
            # Create gen index position lookup
            gen_index_pos_lookup = {}
            for pos, gen_index in enumerate(gen_indexes):
                gen_index_pos_lookup[gen_index.name] = pos

            # Create dependency index set between gen1 and gen2
            gens_uses = itertools.product(
                generations_by_name[var_name], uses_by_name[var_name])
            uses_gens = itertools.product(
                uses_by_name[var_name], generations_by_name[var_name])
            write_read_index_sets.extend(self.generate_dependencies_from_couples(
                itertools.chain(gens_uses, uses_gens),
                indexes,
                index_pos_lookup,
                gen_index_pos_lookup
            ))

        # Eliminate equalities
        index_sets = set(list(write_write_index_sets) +
                         list(write_read_index_sets))
        # Expand sets
        expanded_index_sets = []
        for index_set in index_sets:
            expanded_index_sets.extend(index_set.extract_positives())
        # Convert sets to positive
        positive_sets = tuple(
            map(lambda x: x.to_converted(), expanded_index_sets))
        return positive_sets



from abc import ABC
from datetime import timedelta

import numpy as np
import pymzn

from opoly.modules.minizinc import LAMPORT_SCHEDULER_PATH
from opoly.modules.minizinc.utils import solve_model


class LamportCPScheduler(ABC):

    def schedule(self, deps: np.ndarray) -> (np.ndarray, str):
        if len(deps.shape) != 2:
            return None, "Dependencies must be a matrix!"
        if not issubclass(deps.dtype.type, np.integer):
            return None, "Dependencies must be integers!"
        sol, err = solve_model(
            model=LAMPORT_SCHEDULER_PATH,
            data={
                "r": deps.shape[0],
                "n": deps.shape[1],
                "D": deps.tolist()
            }
        )
        if sol is None:
            return None, err
        return np.array(sol["tau"]), None



from enum import Enum, auto
from abc import ABC, abstractmethod


class VariableOccurrenceType(Enum):
    GENERATION = auto()
    USE = auto()


class VariableOccurrence(ABC):

    def __init__(self, name: str, vtype: VariableOccurrenceType):
        self._vtype = vtype
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    def vtype(self) -> VariableOccurrenceType:
        return self._vtype

    @abstractmethod
    def stringify(self) -> str:
        vtype_str = {
            VariableOccurrenceType.GENERATION: "GEN",
            VariableOccurrenceType.USE: "USE",
        }
        return f"[{vtype_str[self.vtype]}] STM {self.name}"

    def __str__(self):
        return self.stringify()


class SimpleVariableOccurrence(VariableOccurrence):

    def __init__(self, name: str, vtype: VariableOccurrenceType, indexes: tuple[tuple[str, int]]):
        super().__init__(name, vtype)
        self._indexes = indexes

    @property
    def indexes(self) -> tuple[tuple[str, int]]:
        return self._indexes

    @property
    def indexes_names(self) -> tuple[str]:
        return tuple(idx for idx, offset in self._indexes)

    @property
    def offsets(self) -> tuple[int]:
        return tuple(offset for idx, offset in self._indexes)

    def stringify(self) -> str:
        return (f"{super().stringify()}"
                f"{''.join([f'[{idx}{offset:+}]' for idx, offset in self.indexes])}")



import numpy as np
import pytest

from opoly.modules.allocator import LamportCPAllocator

# @pytest.mark.skip(reason="too slow to test every time")
class TestLamportCPAllocator():

    def test_2d_identity(self):
        schedule = [1, 0]
        res, _ = LamportCPAllocator().allocate(np.array(schedule))
        assert res.tolist() == [[1, 0], [0, 1]]

    def test_2d_example1(self):
        schedule = [1, 1]
        res, _ = LamportCPAllocator().allocate(np.array(schedule))
        assert res.tolist() == [[1, 1], [0, 1]]

    def test_2d_example2(self):
        schedule = [2, 1]
        res, _ = LamportCPAllocator().allocate(np.array(schedule))
        assert res.tolist() == [[2, 1], [1, 0]]

    def test_2d_example3(self):
        schedule = [3, 2]
        res, _ = LamportCPAllocator().allocate(np.array(schedule))
        assert res.tolist() == [[3, 2], [1, 1]]

    def test_3d_example4(self):
        schedule = [1, 1, 1]
        res, _ = LamportCPAllocator().allocate(np.array(schedule))
        assert res.tolist() == [
            [1, 1, 1],
            [0, 1, 0],
            [0, 0, 1]
        ]

    def test_3d_example5(self):
        schedule = [2, 1, 1]
        res, _ = LamportCPAllocator().allocate(np.array(schedule))
        assert res.tolist() == [
            [2, 1, 1],
            [1, 0, 0],
            [0, 0, 1]
        ]

    def test_4d(self):
        schedule = [1, 1, 1, 1]
        res, _ = LamportCPAllocator().allocate(np.array(schedule))
        assert res.tolist() == [
            [1, 1, 1, 1],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1]
        ]

    def test_timeout(self):
        schedule = [2, 2]
        res, _ = LamportCPAllocator().allocate(np.array(schedule))
        assert res is None

    def test_unsat(self):
        schedule = [-3, 2]
        res, _ = LamportCPAllocator().allocate(np.array(schedule))
        assert res is None

    def test_not_integer(self):
        schedule = [1.5, 0]
        res, _ = LamportCPAllocator().allocate(np.array(schedule))
        assert res is None



from abc import ABC
from datetime import timedelta

import numpy as np
import pymzn

from opoly.modules.minizinc import LAMPORT_ALLOCATOR_PATH
from opoly.modules.minizinc.utils import solve_model


class LamportCPAllocator(ABC):

    def allocate(self, schedule: np.ndarray) -> (np.ndarray, str):
        if len(schedule.shape) != 1:
            return None, "Schedule must be a one-dimensional vector!"
        sol, err = solve_model(
            model=LAMPORT_ALLOCATOR_PATH,
            data={
                "n": schedule.shape[0],
                "tau": schedule.tolist()
            }
        )
        if sol is None:
            return None, err
        return np.array(sol["T"]), None



import argparse
import pathlib
import logging
import re

import numpy as np

from opoly.modules.parser import PseudocodeForLoopParser
from opoly.modules.checker import LamportForLoopChecker
from opoly.modules.detector import LamportLoopDependenciesDetector
from opoly.modules.scheduler import LamportCPScheduler
from opoly.modules.allocator import LamportCPAllocator
from opoly.modules.scanner import FourierMotzkinScanner
from opoly.modules.generator import CCodeGenerator, PseudoCodeGenerator


def opoly_compile(
    input_file: pathlib.Path,
    output_file: pathlib.Path = None,
    out_format: str = "CCODE",
    verbose: bool = False
):
    logger = logging.getLogger("logger_opoly")
    sh = logging.StreamHandler()
    sh_formatter = logging.Formatter("%(levelname)s - %(message)s")
    sh.setFormatter(sh_formatter)
    logger.addHandler(sh)
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    try:
        logger.debug("Reading input file")
        with open(input_file, "r") as file:
            input_code = file.read()

        logger.debug("Scanning input file")
        omp_poly_dir_regex = re.compile(
            r"#pragma\s+omp\s+parallel\s+poly(?P<code>(.|\s)*?)#pragma\s+end")
        last_pos = 0
        code_list = []
        matches = []
        while True:
            match = omp_poly_dir_regex.search(input_code, pos=last_pos)
            if match is None:
                break
            code_list.append(match.group("code"))
            logger.info(
                f"Found code #{len(code_list)}: " + match.group("code"))
            matches.append(match)
            last_pos = match.end()+1

        if len(matches) == 0:
            logging.warning("No code found. Aborting")
            return

        new_loops = []
        for i, code in enumerate(code_list):
            logger.info(f"Processing code #{i+1}")
            logger.debug("Parsing code")
            loop, err = PseudocodeForLoopParser().parse_for_loop(code)
            if loop is None:
                logger.error(f"Error while parsing code #{i+1}: " + err)
                return
            logger.debug("Checking code")
            ok, err = LamportForLoopChecker().check(loop)
            if not ok:
                logger.error(f"Error while checking code #{i+1}: " + err)
                return
            logger.debug("Detecting dependencies")
            deps = LamportLoopDependenciesDetector().extract_dependencies(loop)
            if len(deps) == 0:
                logger.warning(
                    f"No dependecies found for code #{i+1}. Skipping optimization")
            else:
                deps_np = np.array(list(list(d.converted_values)
                                        for d in deps))
                logger.debug("Scheduling loop")
                schedule, err = LamportCPScheduler().schedule(deps_np)
                if schedule is None:
                    logger.error(f"Error while scheduling loop #{i+1}: " + err)
                    return
                logger.debug("Allocating loop")
                allocation, err = LamportCPAllocator().allocate(schedule)
                if allocation is None:
                    logger.error(f"Error while allocating loop #{i+1}: " + err)
                    return
                logger.debug("Reindexing loop")
                loop = FourierMotzkinScanner().reindex(loop, allocation)
            logger.debug("Generating code")
            generator = CCodeGenerator() if out_format == "CCODE" else PseudoCodeGenerator()
            new_loop_code = generator.generate(loop)
            new_loops.append(new_loop_code)
            logger.debug(f"Loop #{i+1} done")

        # TODO replace old code with new code

    except Exception as ex:
        logger.error(f"An unexpected error as occourred: {ex}")


def main():
    argument_parser = argparse.ArgumentParser(
        description="A simple OpenMP polyhedral compiler for C programs"
    )
    argument_parser.add_argument(
        "file",
        type=pathlib.Path,
        help="the source file containing opoly loops"
    )
    argument_parser.add_argument(
        "-o", "--output",
        type=pathlib.Path,
        metavar="<file>",
        help="place the output into <file>"
    )
    argument_parser.add_argument(
        "-f", "--format",
        type=str,
        choices=["CCODE", "PSEUDO"],
        default="CCODE",
        help="the output format, default CCODE"
    )
    argument_parser.add_argument(
        "-v", "--verbose",
        action="store_true",
        help="make output verbose"
    )
    args = argument_parser.parse_args()
    opoly_compile(
        args.file,
        args.output,
        args.format,
        args.verbose
    )


if __name__ == "__main__":
    main()



import numpy as np

from opoly.modules.scheduler import LamportCPScheduler

class TestlamportCPScheduler():

    def test_2d_identity(self):
        deps = np.array([
            [1,0]
        ])
        sched, _ = LamportCPScheduler().schedule(deps)
        assert sched is not None
        assert sched.tolist() == [1,0]
    
    def test_2d_example1(self):
        deps = np.array([
            [1, 0],
            [0, 1]
        ])
        sched, _ = LamportCPScheduler().schedule(deps)
        assert sched is not None
        assert sched.tolist() == [1,1]
    
    def test_2d_example2(self):
        deps = np.array([
            [1,-1],
            [1, 1],
            [0, 1]
        ])
        sched, _ = LamportCPScheduler().schedule(deps)
        assert sched is not None
        assert sched.tolist() == [2,1]
    
    def test_3d_example4(self):
        deps = np.array([
            [1,0,0],
            [0,1,0],
            [0,0,1],
        ])
        sched, _ = LamportCPScheduler().schedule(deps)
        assert sched is not None
        assert sched.tolist() == [1,1,1]
    
    def test_3d_example5(self):
        deps = np.array([
            [1,0,0],
            [1,-1,0],
            [1,1,0],
            [0,1,0],
            [1,0,-1],
            [1,0,1],
            [0,0,1],
        ])
        sched, _ = LamportCPScheduler().schedule(deps)
        assert sched is not None
        assert sched.tolist() == [2,1,1]


from opoly.expressions import Expression, VariableExpression, ConstantExpression, GroupingExpression
from opoly.statements import ForLoopStatement, AssignmentStatement
from opoly.modules.detector import LamportLoopDependenciesDetector


class TestLamportDetector():

    def test_2d_loop(self):
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(
                left_term=VariableExpression("a", [VariableExpression("j")]),
                right_term=Expression([
                    GroupingExpression([
                        VariableExpression("a", [Expression([
                            VariableExpression("j"), ConstantExpression(1)], "-")]),
                        VariableExpression("a", [VariableExpression("j")]),
                        VariableExpression("a", [Expression([
                            VariableExpression("j"), ConstantExpression(1)], "+")])
                    ], ["+", "+"]),
                    ConstantExpression(3.0)
                ], "/")
            )],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression(
                [VariableExpression("M"), ConstantExpression(1)], ["-"])
        )

        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )

        deps = LamportLoopDependenciesDetector().extract_dependencies(outer_loop)
        assert sorted(map(lambda x: x.converted_values, deps)) == [
            (0, 1), (1, -1), (1, 0), (1, 1)]

    def test_3d_loop(self):
        ajk = VariableExpression("a", [
            VariableExpression("j"),
            VariableExpression("k")
        ])
        ajminus1k = VariableExpression("a", [
            Expression([VariableExpression("j"),
                        ConstantExpression(1)], ["-"]),
            VariableExpression("k")
        ])
        ajminus1kminus1 = VariableExpression("a", [
            Expression([VariableExpression("j"),
                        ConstantExpression(1)], ["-"]),
            Expression([VariableExpression("k"),
                        ConstantExpression(1)], ["-"]),
        ])

        right = Expression([
            ajk, ajminus1k, ajminus1kminus1
        ], ["+", "-"])
        left = ajk

        kloop = ForLoopStatement(
            body=[AssignmentStatement(left, right)],
            index=VariableExpression("k"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("l")
        )

        jloop = ForLoopStatement(
            body=[kloop],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("m")
        )

        iloop = ForLoopStatement(
            body=[jloop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("n")
        )

        deps = LamportLoopDependenciesDetector().extract_dependencies(iloop)
        assert sorted(map(lambda x: x.converted_values, deps)) == [
            (0, 1, 0), (0, 1, 1), (1, -1, -1), (1, -1, 0), (1, 0, 0), (1, 1, 0), (1, 1, 1)]



from opoly.modules.parser import PseudocodeForLoopParser
from opoly.modules.checker import LamportForLoopChecker
from opoly.modules.detector import LamportLoopDependenciesDetector


class TestPseudocodeForLoopParserToLamportDetector():

    def test_1d_loop(self):
        code = "FOR i FROM 0 TO N { STM a[i]=a[i+1]; }"
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        res, _ = LamportForLoopChecker().check(loop)
        deps = LamportLoopDependenciesDetector().extract_dependencies(loop)
        assert sorted(map(lambda x: x.converted_values, deps)) == [(1,)]

    def test_2d_loop(self):
        code = "FOR i FROM 1 TO N-1 { FOR j FROM 2 TO M-1 { STM a[j] = (a[j-1] + a[j] + a[j+1]) / 3.0; } }"
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        res, _ = LamportForLoopChecker().check(loop)
        deps = LamportLoopDependenciesDetector().extract_dependencies(loop)
        assert sorted(map(lambda x: x.converted_values, deps)) == [
            (0, 1), (1, -1), (1, 0), (1, 1)]

    def test_3d_loop(self):
        code = """
            FOR i FROM 1 TO N {
                FOR j FROM 1 TO M-2 {
                    FOR k FROM 1 TO L-2 {
                        STM u[j][k] = (u[j+1][k] + u[j][k+1] + u[j-1][k] + u[j][k-1]) * 0.25;
                    }
                }
            }
        """
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        res, _ = LamportForLoopChecker().check(loop)
        deps = LamportLoopDependenciesDetector().extract_dependencies(loop)
        assert sorted(map(lambda x: x.converted_values, deps)) == [
            (0, 0, 1), (0, 1, 0), (1, -1, 0), (1, 0, -1), (1, 0, 0), (1, 0, 1), (1, 1, 0)]



import pytest
import numpy as np
import sympy as sp

from opoly.expressions import Expression, VariableExpression, ConstantExpression, GroupingExpression
from opoly.statements import AssignmentStatement, ForLoopStatement
from opoly.modules.scanner import reindex, invert_integer_matrix
from opoly.modules.scanner import FourierMotzkinScanner


# @pytest.mark.skip(reason="too slow to test every time")
class TestReindexerFunctions():

    def test_example1(self):
        n, m = sp.symbols("n m", integers=True)
        i, j = sp.symbols("i j", integers=True)
        T = np.array([
            [1, 1],
            [0, 1]
        ], dtype=int)
        idxs = sp.Matrix([[i], [j]])
        ls = sp.Matrix([[1], [1]])
        us = sp.Matrix([[n], [m]])
        bounds = reindex(
            invert_integer_matrix(T),
            idxs,
            ls,
            us
        )
        assert bounds[j] == (sp.Max(1, i - n), sp.Min(m, i - 1))
        assert bounds[i] == (2, n + m)

    def test_example1a(self):
        n, m = sp.symbols("n m", integers=True)
        i, j = sp.symbols("i j", integers=True)
        T = np.array([
            [1, 1],
            [1, 0]
        ], dtype=int)
        idxs = sp.Matrix([[i], [j]])
        ls = sp.Matrix([[1], [1]])
        us = sp.Matrix([[n], [m]])
        bounds = reindex(
            invert_integer_matrix(T),
            idxs,
            ls,
            us
        )
        assert bounds[j] == (sp.Max(1, i - m), sp.Min(n, i - 1))
        assert bounds[i] == (2, n + m)

    def test_example2(self):
        n, m = sp.symbols("n m", integers=True)
        i, j = sp.symbols("i j", integers=True)
        T = np.array([
            [2, 1],
            [1, 0]
        ], dtype=int)
        idxs = sp.Matrix([[i], [j]])
        ls = sp.Matrix([[1], [1]])
        us = sp.Matrix([[n], [m]])
        bounds = reindex(
            invert_integer_matrix(T),
            idxs,
            ls,
            us
        )
        assert bounds[j] == (sp.ceiling(sp.Max(1, sp.together((i - m)/2))),
                             sp.floor(sp.Min(n, sp.together((i - 1)/2))))
        assert bounds[i] == (3, 2*n + m)

    def test_example3(self):
        n, m = sp.symbols("n m", integers=True)
        i, j = sp.symbols("i j", integers=True)
        T = np.array([
            [3, 2],
            [1, 1]
        ], dtype=int)
        idxs = sp.Matrix([[i], [j]])
        ls = sp.Matrix([[1], [1]])
        us = sp.Matrix([[n], [m]])
        bounds = reindex(
            invert_integer_matrix(T),
            idxs,
            ls,
            us
        )
        assert bounds[j] == (sp.ceiling(sp.Max(sp.together((i - n)/2), sp.together((i + 1)/3))),
                             sp.floor(sp.Min(sp.together((i - 1)/2), sp.together((m + i)/3))))
        assert bounds[i] == (5, 3*n + 2*m)

    def test_example4(self):
        l, m, n = sp.symbols("l m n", integers=True)
        i, j, k = sp.symbols("i j k", integers=True)
        T = np.array([
            [1, 1, 1],
            [0, 1, 0],
            [0, 0, 1]
        ], dtype=int)
        idxs = sp.Matrix([[i], [j], [k]])
        ls = sp.Matrix([[1], [1], [1]])
        us = sp.Matrix([[n], [m], [l]])
        bounds = reindex(
            invert_integer_matrix(T),
            idxs,
            ls,
            us
        )
        assert bounds[k] == (sp.Max(1, i - j - n),
                             sp.Min(l, i - j - 1))
        assert bounds[j] == (sp.Max(1, i - n - l),
                             sp.Min(m, i - 2))
        assert bounds[i] == (3, n + m + l)

    def test_example5(self):
        l, m, n = sp.symbols("l m n", integers=True)
        i, j, k = sp.symbols("i j k", integers=True)
        T = np.array([
            [2, 1, 1],
            [1, 0, 0],
            [0, 0, 1]
        ], dtype=int)
        idxs = sp.Matrix([[i], [j], [k]])
        ls = sp.Matrix([[1], [1], [1]])
        us = sp.Matrix([[n], [m], [l]])
        bounds = reindex(
            invert_integer_matrix(T),
            idxs,
            ls,
            us
        )
        assert bounds[k] == (sp.Max(1, i - 2*j - m),
                             sp.Min(l, i - 2*j - 1))
        assert bounds[j] == (sp.ceiling(sp.Max(1, sp.together((i - m - l)/2))),
                             sp.floor(sp.Min(n, sp.together((i - 2)/2))))
        assert bounds[i] == (4, 2*n + m + l)

    def test_example5a(self):
        l, m, n = sp.symbols("l m n", integers=True)
        i, j, k = sp.symbols("i j k", integers=True)
        T = np.array([
            [2, 1, 1],
            [1, 1, 0],
            [0, 0, 1]
        ], dtype=int)
        idxs = sp.Matrix([[i], [j], [k]])
        ls = sp.Matrix([[1], [1], [1]])
        us = sp.Matrix([[n], [m], [l]])
        bounds = reindex(
            invert_integer_matrix(T),
            idxs,
            ls,
            us
        )
        assert bounds[k] == (sp.Max(1, i - 2*j + 1, i - j - n),
                             sp.Min(l, i - 2*j + m, i - j - 1))
        assert bounds[j] == (sp.ceiling(sp.Max(2, i - l - n, sp.together((i - l + 1)/2))),
                             sp.floor(sp.Min(i - 2, m + n, sp.together((i + m - 1)/2))))
        assert bounds[i] == (sp.Max(4, 5 - l, 5 - m),
                             sp.Min(l + m + 2*n, l + 2*m + 2*n - 1, 2*l + m + 2*n - 1))


# @pytest.mark.skip(reason="too slow to test every time")
class TestFourierMotzkinScanner():

    def test_1d_identity(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        allocation = np.array([[1]])
        reindexed_loop = FourierMotzkinScanner().reindex(loop, allocation)
        assert reindexed_loop is not None

    def test_1d_identity2(self):
        loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("i")]),
                                      VariableExpression("b", [VariableExpression("i")]))],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression(
                [VariableExpression("N"), ConstantExpression(1)], "-")
        )
        allocation = np.array([[1]])
        reindexed_loop = FourierMotzkinScanner().reindex(loop, allocation)
        assert reindexed_loop is not None

    def test_2d_identity(self):
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(VariableExpression("a", [VariableExpression("j")]),
                                      VariableExpression("b", [VariableExpression("j")]))],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("M")
        )
        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        allocation = np.array([[1, 0], [0, 1]])
        reindexed_loop = FourierMotzkinScanner().reindex(outer_loop, allocation)
        assert reindexed_loop is not None

    def test_2d_example1(self):
        aij = VariableExpression(
            "a", [VariableExpression("i"), VariableExpression("j")])
        aiminus1j = VariableExpression("a", [Expression([
            VariableExpression("i"),
            ConstantExpression(1)],
            ["-"]),
            VariableExpression("j")
        ])
        aijminus1 = VariableExpression("a", [VariableExpression("i"),
                                             Expression([
                                                 VariableExpression("j"),
                                                 ConstantExpression(1)
                                             ], ["-"])
                                             ])
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(aij, Expression([
                GroupingExpression([aiminus1j, aij, aijminus1], ["+", "+"]),
                ConstantExpression(3.0)], ["/"]))],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression(
                [VariableExpression("M"), ConstantExpression(1)], "-")
        )
        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression(
                [VariableExpression("N"), ConstantExpression(1)], "-")
        )
        allocation = np.array([[1, 1], [0, 1]])
        reindexed_loop = FourierMotzkinScanner().reindex(outer_loop, allocation)
        assert reindexed_loop is not None

    def test_3d_example5(self):
        ujk = VariableExpression("u",[
            VariableExpression("j"),
            VariableExpression("k")
        ])
        ujp1k = VariableExpression("u",[
            Expression([VariableExpression("j"), ConstantExpression(1)], ["+"]),
            VariableExpression("k")
        ])
        ujm1k = VariableExpression("u",[
            Expression([VariableExpression("j"), ConstantExpression(1)], ["-"]),
            VariableExpression("k")
        ])
        ujkp1 = VariableExpression("u",[
            VariableExpression("j"),
            Expression([VariableExpression("k"), ConstantExpression(1)], ["+"])
        ])
        ujkm1 = VariableExpression("u",[
            VariableExpression("j"),
            Expression([VariableExpression("k"), ConstantExpression(1)], ["-"])
        ])
        left = ujk
        right = Expression([
            GroupingExpression([ujp1k, ujm1k, ujkp1, ujkm1], ["+", "+", "+"]),
            ConstantExpression(0.25)
        ], ["/"])

        kloop = ForLoopStatement(
            body=[AssignmentStatement(left, right)],
            index=VariableExpression("k"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression([VariableExpression("L"), ConstantExpression(2)], ["-"])
        )
        jloop = ForLoopStatement(
            body=[kloop],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression([VariableExpression("M"), ConstantExpression(2)], ["-"])
        )
        iloop = ForLoopStatement(
            body=[jloop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=VariableExpression("N")
        )
        allocation = np.array([
            [2, 1, 1],
            [1, 0, 0],
            [0, 0, 1]
        ])
        reindexed_loop = FourierMotzkinScanner().reindex(iloop, allocation)
        assert reindexed_loop is not None
    
    def test_2d_indexes(self):
        aij = VariableExpression(
            "a", [VariableExpression("i"), VariableExpression("j")])
        aiminus1j = VariableExpression("a", [Expression([
            VariableExpression("i"),
            ConstantExpression(1)],
            ["-"]),
            VariableExpression("j")
        ])
        aijminus1 = VariableExpression("a", [VariableExpression("i"),
                                             Expression([
                                                 VariableExpression("j"),
                                                 ConstantExpression(1)
                                             ], ["-"])
                                             ])
        inner_loop = ForLoopStatement(
            body=[AssignmentStatement(aij, Expression([
                GroupingExpression([aiminus1j, aij, aijminus1], ["+", "+"]),
                ConstantExpression(3.0)], ["/"]))],
            index=VariableExpression("j"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression(
                [VariableExpression("i"), ConstantExpression(2)], "+")
        )
        outer_loop = ForLoopStatement(
            body=[inner_loop],
            index=VariableExpression("i"),
            lowerbound=ConstantExpression(1),
            upperbound=Expression(
                [VariableExpression("N"), ConstantExpression(1)], "-")
        )
        allocation = np.array([[1, 1], [0, 1]])
        reindexed_loop = FourierMotzkinScanner().reindex(outer_loop, allocation)
        assert reindexed_loop is not None



import pytest

from opoly.indexes import (
    IndexDescriptorType,
    IndexDescriptor,
    IndexSet
)


class TestIndexDescriptor():

    def test_constant(self):
        desc = IndexDescriptor(
            vdtype=IndexDescriptorType.CONSTANT,
            value=-1
        )
        assert desc.converted_value == -1
        assert str(desc) == "-1"
        assert repr(desc) == "-1"

    def test_any(self):
        desc = IndexDescriptor(
            vdtype=IndexDescriptorType.ANY
        )
        assert desc.converted_value == 1
        assert str(desc) == "*"
        assert repr(desc) == "1"

    def test_positive(self):
        desc = IndexDescriptor(
            vdtype=IndexDescriptorType.POSITIVE
        )
        assert desc.converted_value == 1
        assert str(desc) == "+"
        assert repr(desc) == "1"

    def test_wrong_constant(self):
        with pytest.raises(ValueError):
            _ = IndexDescriptor(
                vdtype=IndexDescriptorType.CONSTANT
            )

    def test_wrong_non_constant(self):
        with pytest.raises(ValueError):
            _ = IndexDescriptor(
                vdtype=IndexDescriptorType.POSITIVE,
                value=100
            )


class TestIndexSet():

    def test_simple(self):
        vset = IndexSet(
            descriptors=(
                IndexDescriptor(IndexDescriptorType.CONSTANT, 1),
                IndexDescriptor(IndexDescriptorType.CONSTANT, -1)
            )
        )
        assert vset.converted_values == (1, -1)
        assert str(vset) == "(1,-1)"
        assert repr(vset) == "(1,-1)"
        assert list([str(s) for s in vset.extract_positives()]) == [
            "(1,-1)"
        ]

    def test_any(self):
        vset = IndexSet(
            descriptors=(
                IndexDescriptor(IndexDescriptorType.ANY),
                IndexDescriptor(IndexDescriptorType.CONSTANT, 1)
            )
        )
        assert vset.converted_values == (1, 1)
        assert str(vset) == "(*,1)"
        assert repr(vset) == "(1,1)"
        assert list([str(s) for s in vset.extract_positives()]) == [
            "(0,1)",
            "(+,1)"
        ]

    def test_complex(self):
        vset = IndexSet(
            descriptors=(
                IndexDescriptor(IndexDescriptorType.ANY),
                IndexDescriptor(IndexDescriptorType.CONSTANT, 1),
                IndexDescriptor(IndexDescriptorType.POSITIVE),
                IndexDescriptor(IndexDescriptorType.CONSTANT, -1)
            )
        )
        assert vset.converted_values == (1, 1, 1, -1)
        assert str(vset) == "(*,1,+,-1)"
        assert repr(vset) == "(1,1,1,-1)"
        assert list([str(s) for s in vset.extract_positives()]) == [
            "(0,1,+,-1)",
            "(+,1,+,-1)"
        ]

    def test_successive_any_alone(self):
        vset = IndexSet(
            descriptors=(
                IndexDescriptor(IndexDescriptorType.ANY),
                IndexDescriptor(IndexDescriptorType.ANY)
            )
        )
        assert vset.converted_values == (1, 1)
        assert str(vset) == "(*,*)"
        assert repr(vset) == "(1,1)"
        assert list([str(s) for s in vset.extract_positives()]) == [
            "(0,+)",
            "(+,0)",
            "(+,+)"
        ]

    def test_successive_any_with_positive_constant(self):
        vset = IndexSet(
            descriptors=(
                IndexDescriptor(IndexDescriptorType.ANY),
                IndexDescriptor(IndexDescriptorType.ANY),
                IndexDescriptor(IndexDescriptorType.CONSTANT, 1),
            )
        )
        assert vset.converted_values == (1, 1, 1)
        assert str(vset) == "(*,*,1)"
        assert repr(vset) == "(1,1,1)"
        assert list([str(s) for s in vset.extract_positives()]) == [
            "(0,0,1)",
            "(0,+,1)",
            "(+,0,1)",
            "(+,+,1)"
        ]

    def test_successive_any_with_negative_constant(self):
        vset = IndexSet(
            descriptors=(
                IndexDescriptor(IndexDescriptorType.ANY),
                IndexDescriptor(IndexDescriptorType.ANY),
                IndexDescriptor(IndexDescriptorType.CONSTANT, -1),
            )
        )
        assert vset.converted_values == (1, 1, -1)
        assert str(vset) == "(*,*,-1)"
        assert repr(vset) == "(1,1,-1)"
        assert list([str(s) for s in vset.extract_positives()]) == [
            "(0,+,-1)",
            "(+,0,-1)",
            "(+,+,-1)"
        ]

    def test_multiple_any(self):
        vset = IndexSet(
            descriptors=(
                IndexDescriptor(IndexDescriptorType.ANY),
                IndexDescriptor(IndexDescriptorType.CONSTANT, -1),
                IndexDescriptor(IndexDescriptorType.ANY),
            )
        )
        assert vset.converted_values == (1, -1, 1)
        assert str(vset) == "(*,-1,*)"
        assert repr(vset) == "(1,-1,1)"
        assert list([str(s) for s in vset.extract_positives()]) == [
            "(+,-1,0)",
            "(+,-1,+)"
        ]

    def test_positive(self):
        vset = IndexSet(
            descriptors=(
                IndexDescriptor(IndexDescriptorType.ANY),
                IndexDescriptor(IndexDescriptorType.CONSTANT, -1),
                IndexDescriptor(IndexDescriptorType.CONSTANT, 1)
            )
        )
        assert vset.is_lexico_positive()

    def test_not_positive(self):
        vset = IndexSet(
            descriptors=(
                IndexDescriptor(IndexDescriptorType.CONSTANT, -1),
                IndexDescriptor(IndexDescriptorType.CONSTANT, 1)
            )
        )
        assert not vset.is_lexico_positive()

    def test_zeros(self):
        vset = IndexSet(
            descriptors=(
                IndexDescriptor(IndexDescriptorType.CONSTANT, 0),
                IndexDescriptor(IndexDescriptorType.CONSTANT, 0)
            )
        )
        assert vset.converted_values == (0, 0)
        assert str(vset) == "(0,0)"
        assert repr(vset) == "(0,0)"
        assert not vset.is_lexico_positive()

    def test_no_descriptors(self):
        with pytest.raises(ValueError):
            _ = IndexSet(descriptors=())



from opoly.modules.parser import PseudocodeForLoopParser
from opoly.modules.checker import LamportForLoopChecker


class TestPseudocodeForLoopParserToLamportChecker():

    def test_1d_loop(self):
        code = "FOR i FROM 0 TO N { STM a[i]=a[i+1]; }"
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        res, _ = LamportForLoopChecker().check(loop)
        assert res

    def test_2d_loop(self):
        code = "FOR i FROM 1 TO N-1 { FOR j FROM 2 TO M-1 { STM a[j] = (a[j-1] + a[j] + a[j+1]) / 3.0; } }"
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        res, _ = LamportForLoopChecker().check(loop)
        assert res

    def test_3d_loop(self):
        code = """
            FOR i FROM 1 TO N {
                FOR j FROM 2 TO M-2 {
                    FOR k FROM 2 TO L-2 {
                        STM a[j][k] = (a[j-1][k-1] + a[j][k] + a[j+1][k+1]) / 3.0;
                        STM a[j+2][k+2] = a[j][k];
                    }
                }
            }
        """
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        res, _ = LamportForLoopChecker().check(loop)
        assert res

    def test_not_perfectly_nested_loop(self):
        code = """
        FOR i FROM 0 TO N {
            STM a[i][0] = 0;
            FOR j FROM 1 TO M-1 {
                STM a[i][j] = a[i][j-1] + a[i][j+1];
                STM a[j][j] = 0;
            }
            STM a[N][M] = a[0][0];
        }
        """
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        res, err = LamportForLoopChecker().check(loop)
        assert not res
        assert err == "Not perfectly nested loop!"

    def test_not_plain_loop(self):
        code = """
        FOR i FROM 0 TO N {
            FOR j FROM i TO M {
                STM a = b;
            }
        }
        """
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        res, err = LamportForLoopChecker().check(loop)
        assert not res
        assert err == "Not plain loop!"



import numpy as np
import sympy as sp
from sympy.solvers.inequalities import reduce_inequalities

from opoly.expressions import ConstantExpression, VariableExpression
from opoly.statements import ForLoopStatement, DeclarationStatement
from opoly.modules.checker import (
    extract_loop_indexes,
    extract_loop_bounds,
    get_simple_variable_sum_and_constant,
    get_inner_loop_statements
)
from opoly.modules.parser import parse_expression

def reduce_ineq(
    ineq: sp.core.relational._Inequality,
    var: sp.core.symbol.Symbol
) -> sp.core.relational._Inequality:
    interval = reduce_inequalities(ineq, [var])
    args = interval.args if isinstance(interval, sp.And) else [interval]
    for inter in args:
        if not inter.lhs.has(var):
            inter = inter.canonical
        if not (inter.has(sp.oo) or inter.has(-sp.oo)):
            return inter


def fourier_motzkin_eliminate_var(
    inequalites: tuple[sp.core.relational._Inequality],
    var: sp.core.symbol.Symbol,
    all_vars: tuple[sp.core.symbol.Symbol]
) -> tuple[sp.core.relational._Inequality]:
    less_exprs = []
    great_exprs = []
    zero_ineqs = []
    for ineq in inequalites:
        if var in ineq.free_symbols:
            ineq = reduce_ineq(ineq, var)
            if isinstance(ineq, sp.core.relational.GreaterThan):
                great_exprs.append(ineq.rhs)
            else:
                less_exprs.append(ineq.rhs)
        else:
            zero_ineqs.append(ineq)
    new_ineqs = []
    for ge in great_exprs:
        for le in less_exprs:
            diff = ge - le
            if any(diff.has(v) for v in all_vars):
                new_ineqs.append(ge <= le)
    return tuple(zero_ineqs + new_ineqs)


def extract_var_bounds(
    ineqs: tuple[sp.core.relational._Inequality],
    var: sp.core.symbol.Symbol
) -> (tuple[sp.core.expr.Expr], tuple[sp.core.expr.Expr]):
    lower_bounds = []
    upper_bounds = []
    for ineq in ineqs:
        if var in ineq.free_symbols:
            ineq = reduce_ineq(ineq, var)
            if ineq.has(var):
                bound = sp.together(ineq.rhs)
                if isinstance(ineq, sp.core.relational.GreaterThan):
                    lower_bounds.append(bound)
                else:
                    upper_bounds.append(bound)
    return (tuple(lower_bounds), tuple(upper_bounds))


def fourier_motzkin(
    ineqs: tuple[sp.core.relational._Inequality],
    all_vars: tuple[sp.core.symbol.Symbol],
    last_index: int = 0
) -> dict[sp.core.symbol.Symbol, (tuple[sp.core.expr.Expr], tuple[sp.core.expr.Expr])]:
    dim = len(all_vars)
    bounds_dict = {}
    curr_index = dim - 1

    while curr_index > last_index:
        curr_var = all_vars[curr_index]
        bounds_dict[curr_var] = extract_var_bounds(ineqs, curr_var)
        ineqs = fourier_motzkin_eliminate_var(ineqs, curr_var, all_vars)
        curr_index -= 1

    last_var = all_vars[last_index]
    bounds_dict[last_var] = extract_var_bounds(ineqs, last_var)
    return bounds_dict


def enclose_bounds(
    bounds_dict: dict[sp.core.symbol.Symbol,
                      (tuple[sp.core.expr.Expr], tuple[sp.core.expr.Expr])]
) -> dict[sp.core.symbol.Symbol, (tuple[sp.core.expr.Expr], tuple[sp.core.expr.Expr])]:
    enclosed_bounds_dict = {}
    for var, bounds in bounds_dict.items():
        lowers, uppers = bounds
        to_ceil = any(map(lambda l: isinstance(l, sp.core.mul.Mul) and
                                    any(map(lambda ll: ll.is_rational and not ll.is_integer, l.as_two_terms())),
                          lowers))
        to_floor = any(map(lambda u: isinstance(u, sp.core.mul.Mul) and
                                     any(map(lambda uu: uu.is_rational and not uu.is_integer, u.as_two_terms())),
                           uppers))
        enclosed_bounds_dict[var] = (
            sp.ceiling(sp.Max(*lowers)) if to_ceil else sp.Max(*lowers),
            sp.floor(sp.Min(*uppers)) if to_floor else sp.Min(*uppers)
        )
    return enclosed_bounds_dict


def reindex(T_inv: np.array, x: sp.Matrix, ls: sp.Matrix, us: sp.Matrix):
    system = T_inv * x
    ineqs = []
    for i, t in enumerate(system):
        ineqs.append(t >= ls[i])
        ineqs.append(t <= us[i])
    all_vars = tuple(xx for xx in x)
    bounds = fourier_motzkin(ineqs, all_vars)
    return enclose_bounds(bounds)

def invert_integer_matrix(mat: np.ndarray):
    return np.linalg.inv(mat).round().astype(int)

class FourierMotzkinScanner():

    def generate_reversed_index_declarations(self, old_indexes, new_indexes, inverted_allocation):
        declarations = []
        inverted_indexes = inverted_allocation * new_indexes
        for i, old_idx in enumerate(old_indexes):
            declarations.append(DeclarationStatement(
                var_type="int",
                variable=old_idx,
                initialization=parse_expression(sp.ccode(inverted_indexes[i]))[0]
            ))
        return tuple(declarations)

    def extract_bounds(self, bounds, old_indexes, inv_indexes):
        res = []
        for b in bounds:
            if b.is_constant():
                res.append(b.value)
            elif b.is_variable():
                res.append(sp.var(b.name, integer=True))
            else:
                v, c = get_simple_variable_sum_and_constant(b)
                found = False
                for i, oi in enumerate(old_indexes):
                    if v.name == oi.name:
                        found = True
                        res.append(inv_indexes[i] + c.value)
                        break
                if not found:
                    res.append(sp.var(v.name, integer=True) + c.value)
        return sp.Matrix(res)

    def reindex(self, loop: ForLoopStatement, allocation: np.ndarray, separate_bounds: bool = False) -> ForLoopStatement:
        old_indexes = extract_loop_indexes(loop)
        new_indexes = list(VariableExpression("new_" + i.name) for i in old_indexes) #pylint: disable=no-member
        indexes = sp.Matrix(list(
            [sp.var(i.name, integer=True) for i in new_indexes]
        ))
        inverted_allocation = invert_integer_matrix(allocation)
        lower_bounds, upper_bounds = list(zip(*extract_loop_bounds(loop)))
        inv_indexes = inverted_allocation * indexes
        ls = self.extract_bounds(lower_bounds, old_indexes, inv_indexes)
        us = self.extract_bounds(upper_bounds, old_indexes, inv_indexes)
        new_bounds = reindex(
            inverted_allocation,
            indexes,
            ls,
            us
        )
        statements = get_inner_loop_statements(loop)

        last_loop = None
        last_bounds = None
        for i, new_idx in reversed(list(enumerate(indexes))):
            lb, _ = parse_expression(sp.ccode(new_bounds[new_idx][0]))
            ub, _ = parse_expression(sp.ccode(new_bounds[new_idx][1]))

            if last_loop is None:
                reverse_declarations = self.generate_reversed_index_declarations(
                    old_indexes,
                    indexes,
                    inverted_allocation
                )
                body = list(reverse_declarations) + list(statements)
            else:
                body = [last_loop]
            
            if last_bounds is not None:
                body = last_bounds + body
            
            if i >= 1 and separate_bounds:
                lb_decl = DeclarationStatement(
                    var_type="int",
                    variable=VariableExpression(new_indexes[i].name + "_lb"),
                    initialization=lb
                )
                ub_decl = DeclarationStatement(
                    var_type="int",
                    variable=VariableExpression(new_indexes[i].name + "_ub"),
                    initialization=ub
                )
                lb = lb_decl.variable
                ub = ub_decl.variable
                last_bounds = [lb_decl, ub_decl]

            
            last_loop = ForLoopStatement(
                body=tuple(body),
                index=VariableExpression(repr(new_idx)),
                lowerbound=lb,
                upperbound=ub,
                is_parallel=(i == 1)
            )
        return last_loop



import argparse
import pathlib
import logging
import re

import numpy as np

from opoly.modules.parser import PseudocodeForLoopParser
from opoly.modules.checker import LamportForLoopChecker
from opoly.modules.detector import LamportLoopDependenciesDetector
from opoly.modules.scheduler import LamportCPScheduler
from opoly.modules.allocator import LamportCPAllocator
from opoly.modules.scanner import FourierMotzkinScanner
from opoly.modules.generator import CCodeGenerator, PseudoCodeGenerator


def opoly(
    input_file: pathlib.Path,
    output_file: pathlib.Path = None,
    out_format: str = "CCODE",
    verbose: bool = False
):
    logger = logging.getLogger("logger_opoly")
    sh = logging.StreamHandler()
    sh_formatter = logging.Formatter("%(levelname)s - %(message)s")
    sh.setFormatter(sh_formatter)
    logger.addHandler(sh)
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    try:
        logger.debug("Reading input file")
        with open(input_file, "r") as file:
            code = file.read()

        logger.debug("Parsing code")
        loop, err = PseudocodeForLoopParser().parse_for_loop(code)
        if loop is None:
            logger.error("Error while parsing code: " + err)
            return

        logger.debug("Checking code")
        ok, err = LamportForLoopChecker().check(loop)
        if not ok:
            logger.error("Error while checking code: " + err)
            return

        logger.debug("Detecting code dependencies")
        deps = LamportLoopDependenciesDetector().extract_dependencies(loop)
        if len(deps) == 0:
            logger.warning(
                "No dependecies found in code. Skipping optimization")
        else:
            deps_np = np.array(list(list(d.converted_values)
                                    for d in deps))

            logger.debug("Scheduling loop")
            schedule, err = LamportCPScheduler().schedule(deps_np)
            if schedule is None:
                logger.error("Error while scheduling loop: " + err)
                return

            logger.debug("Allocating loop")
            allocation, err = LamportCPAllocator().allocate(schedule)
            if allocation is None:
                logger.error("Error while allocating loop: " + err)
                return

            logger.debug("Reindexing loop")
            loop = FourierMotzkinScanner().reindex(loop, allocation, separate_bounds=out_format == "CCODE")
        logger.debug("Generating code")
        generator = CCodeGenerator() if out_format == "CCODE" else PseudoCodeGenerator()
        new_code = generator.generate(loop)

        if output_file is None:
            print(new_code)
        else:
            logger.debug("Writing output file")
            with open(output_file, "w") as file:
                file.write(new_code)

    except Exception as ex:
        logger.error(f"An unexpected error as occourred: {ex}")


def main():
    argument_parser = argparse.ArgumentParser(
        description="A simple OpenMP polyhedral compiler for C programs"
    )
    argument_parser.add_argument(
        "file",
        type=pathlib.Path,
        help="the source file containing opoly loops"
    )
    argument_parser.add_argument(
        "-o", "--output",
        type=pathlib.Path,
        metavar="<file>",
        help="place the output into <file>"
    )
    argument_parser.add_argument(
        "-f", "--format",
        type=str,
        choices=["CCODE", "PSEUDO"],
        default="CCODE",
        help="the output format, default CCODE"
    )
    argument_parser.add_argument(
        "-v", "--verbose",
        action="store_true",
        help="make output verbose"
    )
    args = argument_parser.parse_args()
    opoly(
        args.file,
        args.output,
        args.format,
        args.verbose
    )


if __name__ == "__main__":
    main()



import pytest

import numpy as np

from opoly.modules.parser import PseudocodeForLoopParser
from opoly.modules.checker import LamportForLoopChecker
from opoly.modules.detector import LamportLoopDependenciesDetector
from opoly.modules.scheduler import LamportCPScheduler
from opoly.modules.allocator import LamportCPAllocator
from opoly.modules.scanner import FourierMotzkinScanner
from opoly.modules.generator import PseudoCodeGenerator, CCodeGenerator


class TestPseudocodeForLoopParserToPseudoCodeGenerator():

    def _pipeline_parser_scheduler(self, code: str) -> np.ndarray:
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        _, _ = LamportForLoopChecker().check(loop)
        deps = LamportLoopDependenciesDetector().extract_dependencies(loop)
        deps_np = np.array(list(list(d.converted_values) for d in deps))
        schedule, _ = LamportCPScheduler().schedule(deps_np)
        allocation, _ = LamportCPAllocator().allocate(schedule)
        reindexed_loop = FourierMotzkinScanner().reindex(loop, allocation)
        parallel_code = PseudoCodeGenerator().generate(reindexed_loop)
        return parallel_code

    def test_1d_loop(self):
        code = "FOR i FROM 0 TO N { STM a[i]=a[i+1]; }"
        parallel_code = self._pipeline_parser_scheduler(code)
        assert parallel_code == """FOR new_i FROM 0 TO N STEP 1 {
    VAR i = new_i;
    STM a[i] = a[i + 1];
}"""

    def test_2d_loop_example1(self):
        code = "FOR i FROM 1 TO N-1 { FOR j FROM 1 TO M-1 { STM a[i][j] = (a[i-1][j] + a[i][j] + a[i][j-1]) / 3.0; } }"
        parallel_code = self._pipeline_parser_scheduler(code)
        print(parallel_code)
        assert parallel_code == """FOR new_i FROM 2 TO M + N - 2 STEP 1 {
    FOR CONC new_j FROM fmax(1, -N + new_i + 1) TO fmin(M - 1, new_i - 1) STEP 1 {
        VAR i = new_i - new_j;
        VAR j = new_j;
        STM a[i][j] = (a[i - 1][j] + a[i][j] + a[i][j - 1]) / 3.0;
    }
}"""

    def test_2d_loop_example2(self):
        code = "FOR i FROM 1 TO N-1 { FOR j FROM 1 TO M-1 { STM a[j] = (a[j-1] + a[j] + a[j+1]) / 3.0; } }"
        parallel_code = self._pipeline_parser_scheduler(code)
        print(parallel_code)
        assert parallel_code == """FOR new_i FROM 3 TO M + 2 * N - 3 STEP 1 {
    FOR CONC new_j FROM ceil(fmax(1, (1.0 / 2.0) * (-M + new_i + 1))) TO floor(fmin(N - 1, (1.0 / 2.0) * (new_i - 1))) STEP 1 {
        VAR i = new_j;
        VAR j = new_i - 2 * new_j;
        STM a[j] = (a[j - 1] + a[j] + a[j + 1]) / 3.0;
    }
}"""

    def test_3d_loop_example5(self):
        code = """
        FOR i FROM 1 TO N-1 {
            FOR j FROM 1 TO M-2 {
                FOR k FROM 1 TO L-2 {
                    STM u[j][k] = (u[j+1][k] + u[j][k+1] + u[j-1][k] + u[j][k-1]) * 0.25;
                }
            }
        }
        """
        parallel_code = self._pipeline_parser_scheduler(code)
        assert parallel_code == """FOR new_i FROM 4 TO L + M + 2 * N - 6 STEP 1 {
    FOR CONC new_j FROM ceil(fmax(1, (1.0 / 2.0) * (-L - M + new_i + 4))) TO floor(fmin(N - 1, (1.0 / 2.0) * (new_i - 2))) STEP 1 {
        FOR new_k FROM fmax(1, -M + new_i - 2 * new_j + 2) TO fmin(L - 2, new_i - 2 * new_j - 1) STEP 1 {
            VAR i = new_j;
            VAR j = new_i - 2 * new_j - new_k;
            VAR k = new_k;
            STM u[j][k] = (u[j + 1][k] + u[j][k + 1] + u[j - 1][k] + u[j][k - 1]) * 0.25;
        }
    }
}"""

# @pytest.mark.skip(reason="too slow to test every time")
class TestPseudocodeForLoopParserToCCodeGenerator():

    def _pipeline_parser_scheduler(self, code: str) -> np.ndarray:
        loop, _ = PseudocodeForLoopParser().parse_for_loop(code)
        _, _ = LamportForLoopChecker().check(loop)
        deps = LamportLoopDependenciesDetector().extract_dependencies(loop)
        deps_np = np.array(list(list(d.converted_values) for d in deps))
        schedule, _ = LamportCPScheduler().schedule(deps_np)
        allocation, _ = LamportCPAllocator().allocate(schedule)
        reindexed_loop = FourierMotzkinScanner().reindex(loop, allocation)
        parallel_code = CCodeGenerator().generate(reindexed_loop)
        return parallel_code

    def test_1d_loop(self):
        code = "FOR i FROM 0 TO N { STM a[i]=a[i+1]; }"
        parallel_code = self._pipeline_parser_scheduler(code)
        print(parallel_code)
        assert parallel_code == """for(int new_i = 0; new_i <= N; new_i++) {
    int i = new_i;
    a[i] = a[i + 1];
}"""

    def test_2d_loop_example1(self):
        code = "FOR i FROM 1 TO N-1 { FOR j FROM 1 TO M-1 { STM a[i][j] = (a[i-1][j] + a[i][j] + a[i][j-1]) / 3.0; } }"
        parallel_code = self._pipeline_parser_scheduler(code)
        print(parallel_code)
        assert parallel_code == """for(int new_i = 2; new_i <= M + N - 2; new_i++) {
    #pragma omp parallel for
    for(int new_j = fmax(1, -N + new_i + 1); new_j <= fmin(M - 1, new_i - 1); new_j++) {
        int i = new_i - new_j;
        int j = new_j;
        a[i][j] = (a[i - 1][j] + a[i][j] + a[i][j - 1]) / 3.0;
    }
}"""

    def test_2d_loop_example2(self):
        code = "FOR i FROM 1 TO N-1 { FOR j FROM 1 TO M-1 { STM a[j] = (a[j-1] + a[j] + a[j+1]) / 3.0; } }"
        parallel_code = self._pipeline_parser_scheduler(code)
        print(parallel_code)
        assert parallel_code == """for(int new_i = 3; new_i <= M + 2 * N - 3; new_i++) {
    #pragma omp parallel for
    for(int new_j = ceil(fmax(1, (1.0 / 2.0) * (-M + new_i + 1))); new_j <= floor(fmin(N - 1, (1.0 / 2.0) * (new_i - 1))); new_j++) {
        int i = new_j;
        int j = new_i - 2 * new_j;
        a[j] = (a[j - 1] + a[j] + a[j + 1]) / 3.0;
    }
}"""

    def test_3d_loop_example5(self):
        code = """
        FOR i FROM 1 TO N-1 {
            FOR j FROM 1 TO M-2 {
                FOR k FROM 1 TO L-2 {
                    STM u[j][k] = (u[j+1][k] + u[j][k+1] + u[j-1][k] + u[j][k-1]) * 0.25;
                }
            }
        }
        """
        parallel_code = self._pipeline_parser_scheduler(code)
        print(parallel_code)
        assert parallel_code == """for(int new_i = 4; new_i <= L + M + 2 * N - 6; new_i++) {
    #pragma omp parallel for
    for(int new_j = ceil(fmax(1, (1.0 / 2.0) * (-L - M + new_i + 4))); new_j <= floor(fmin(N - 1, (1.0 / 2.0) * (new_i - 2))); new_j++) {
        for(int new_k = fmax(1, -M + new_i - 2 * new_j + 2); new_k <= fmin(L - 2, new_i - 2 * new_j - 1); new_k++) {
            int i = new_j;
            int j = new_i - 2 * new_j - new_k;
            int k = new_k;
            u[j][k] = (u[j + 1][k] + u[j][k + 1] + u[j - 1][k] + u[j][k - 1]) * 0.25;
        }
    }
}"""

    def test_3d_loop_lamport(self):
        code = """
            FOR i FROM 1 TO L {
                FOR j FROM 2 TO M {
                    FOR k FROM 2 TO N {
                        STM u[j][k] = (u[j+1][k] + u[j][k+1] + u[j-1][k] + u[j][k-1]) * 0.25;
                    }
                }
            }
            """
        parallel_code = self._pipeline_parser_scheduler(code)
        assert parallel_code == """for(int new_i = 6; new_i <= 2 * L + M + N; new_i++) {
    #pragma omp parallel for
    for(int new_j = ceil(fmax(1, (1.0 / 2.0) * (-M - N + new_i))); new_j <= floor(fmin(L, (1.0 / 2.0) * (new_i - 4))); new_j++) {
        for(int new_k = fmax(2, -M + new_i - 2 * new_j); new_k <= fmin(N, new_i - 2 * new_j - 2); new_k++) {
            int i = new_j;
            int j = new_i - 2 * new_j - new_k;
            int k = new_k;
            u[j][k] = (u[j + 1][k] + u[j][k + 1] + u[j - 1][k] + u[j][k - 1]) * 0.25;
        }
    }
}"""

#     def test_separate_bounds(self):
#         code = """
#             FOR i FROM 1 TO L {
#                 FOR j FROM 2 TO M {
#                     FOR k FROM 2 TO N {
#                         STM u[j][k] = (u[j+1][k] + u[j][k+1] + u[j-1][k] + u[j][k-1]) * 0.25;
#                     }
#                 }
#             }
#             """
#         parallel_code = self._pipeline_parser_scheduler(code)
#         assert parallel_code == """for(int new_i = 6; new_i <= 2 * L + M + N; new_i++) {
#     int new_j_lb = ceil(fmax(1, (1.0 / 2.0) * (-M - N + new_i)));
#     int new_j_ub = floor(fmin(L, (1.0 / 2.0) * (new_i - 4)));
#     #pragma omp parallel for
#     for(int new_j = new_j_lb; new_j <= new_j_ub; new_j++) {
#         for(int new_k = fmax(2, -M + new_i - 2 * new_j); new_k <= fmin(N, new_i - 2 * new_j - 2); new_k++) {
#             int i = new_j;
#             int j = new_i - 2 * new_j - new_k;
#             int k = new_k;
#             u[j][k] = (u[j + 1][k] + u[j][k + 1] + u[j - 1][k] + u[j][k - 1]) * 0.25;
#         }
#     }
# }"""




from abc import ABC, abstractmethod
import re

from opoly.expressions import (
    Expression, ConstantExpression, VariableExpression, GroupingExpression, FunctionExpression, UnaryExpression)
from opoly.statements import Statement, ForLoopStatement, AssignmentStatement

FUNCTION_EXPRESSION_REGEX = re.compile(r"^(?P<name>\w+)\((?P<terms>.*)\)")
FUNCTION_TERMS_EXPRESSION_REGEX = re.compile(r"^(?P<term>[^,\s].*?)(,|$)")
UNARY_OPERATORS = [
    "-"
]
BINARY_OPERATORS = [
    "+", "-", "*", "/"
]


def isfloat(s: str) -> bool:
    try:
        _ = float(s)
    except (TypeError, ValueError):
        return False
    else:
        return True


def isint(s: str):
    try:
        _ = int(s)
    except (TypeError, ValueError):
        return False
    else:
        return True

# TODO fix every return value with Optional[]


def parse_constant_expression(code: str) -> (ConstantExpression, str):
    if len(code) == 0:
        return None, "Expected constant expression"
    i = 0
    while i < len(code) and (code[i].isnumeric() or code[i] == "."):
        i += 1
    res_str = code[:i]
    if isint(res_str):
        res = int(res_str)
    elif isfloat(res_str):
        res = float(res_str)
    else:
        return None, "Expected constant expression"
    return ConstantExpression(res), code[i:]


def parse_variable_expression(code: str) -> (VariableExpression, str):
    if len(code) == 0:
        return None, "Expected variable expression"
    i = 0
    while i < len(code) and (code[i].isalpha() or code[i] == "_"):
        i += 1
    name = code[:i]
    if name == "":
        return None, "Expected variable expression"
    code = code[i:]
    indexes = []
    while len(code) > 0 and code[0] == "[":
        var_index, rem_code = parse_expression(code[1:].strip(), term_char="]")
        if var_index is None:
            return None, rem_code
        rem_code = rem_code.strip()
        if len(rem_code) == 0 or rem_code[0] != "]":
            return None, "Expected ]"
        indexes.append(var_index)
        code = rem_code[1:]
    return VariableExpression(name, indexes), code


def parse_operator(code: str) -> (str, str):
    if len(code) == 0:
        return None, "Expected operator"
    if code[0] not in BINARY_OPERATORS:
        return None, "Unsupported operator"
    return code[0], code[1:]


def parse_grouping_expression(code: str) -> (GroupingExpression, str):
    if len(code) == 0 or code[0] != "(":
        return None, "Expected ("
    expr, rem_code = parse_expression(code[1:], term_char=")")
    if expr is None:
        return None, rem_code
    rem_code = rem_code.strip()
    if len(rem_code) == 0 or rem_code[0] != ")":
        return None, "Expected )"
    return GroupingExpression([expr], []), rem_code[1:]


def parse_simple_function_expression(code: str) -> (FunctionExpression, str):
    match = FUNCTION_EXPRESSION_REGEX.match(code)
    if not match:
        return None, "Expected function expression"
    name = match.group("name")
    terms_code = match.group("terms").strip()
    terms = []
    while (term_match := FUNCTION_TERMS_EXPRESSION_REGEX.match(terms_code)) is not None:
        term, err = parse_expression(term_match.group("term").strip())
        if term is None:
            return None, err
        terms_code = terms_code[term_match.span()[1]:].strip()
        terms.append(term)
    if len(terms_code) != 0:
        return None, "Function expression not well formatted"
    return FunctionExpression(name, args=terms), code[match.span()[1]:]

def parse_function_expression(code: str) -> (FunctionExpression, str):
    if len(code) == 0:
        return None, "Expected function expression"
    i = 0
    while i < len(code) and (code[i].isalpha() or code[i] == "_"):
        i += 1
    name = code[:i]
    if name == "":
        return None, "Expected function name"
    code = code[i:]
    if len(code) == 0 or code[0] != "(":
        return None, "Expected ("
    expressions = []
    while True:
        expr, rem_code = parse_expression(code[1:], term_char=",")
        if expr is None:
            break
        rem_code = rem_code.strip()
        if len(rem_code) == 0 or rem_code[0] != ",":
            return None, "Expected ,"
        code = rem_code
        expressions.append(expr)
        
    last_expr, rem_code = parse_expression(code[1:], term_char=")")
    if last_expr is None:
        return None, rem_code
    expressions.append(last_expr)
    rem_code = rem_code.strip()
    if len(rem_code) == 0 or rem_code[0] != ")":
        return None, "Expected )"
    return FunctionExpression(name, tuple(expressions)), rem_code[1:]

def parse_unary_expression(code: str) -> (UnaryExpression, str):
    if len(code) == 0:
        return None, "Expected unary expression"
    if code[0] not in UNARY_OPERATORS:
        return None, "Unsupported unary operator"
    if parse_unary_expression(code[1:].strip())[0] is not None:
        return None, "Unsupported unary operator"
    term, rem_code = parse_single_expression(code[1:].strip())
    if term is None:
        return None, rem_code
    return UnaryExpression(term, code[0]), rem_code

def parse_single_expression(code: str) -> (Expression, str):
    code = code.strip()
    # Check for expression
    if len(code) == 0:
        return None, "Expected expression"

    # Check different subexpression terms
    if FUNCTION_EXPRESSION_REGEX.match(code) is not None:
        next_term, code = parse_function_expression(code)
    elif code[0].isalpha():
        next_term, code = parse_variable_expression(code)
    elif code[0].isnumeric():
        next_term, code = parse_constant_expression(code)
    elif code[0] == "(":
        next_term, code = parse_grouping_expression(code)
    elif code[0] in UNARY_OPERATORS:
        next_term, code = parse_unary_expression(code)
    else:
        return None, "Unsupported expression term"
    return next_term, code

def parse_expression(code: str, term_char=None) -> (Expression, str):
    terms = []
    operators = []
    while True:
        # Parse next expression
        next_term, code = parse_single_expression(code)
        if next_term is None:
            return None, code
        terms.append(next_term)

        # Check for end
        code = code.strip()
        if len(code) == 0:
            break

        # Check for operator
        next_operator, rem_code = parse_operator(code)
        if next_operator is not None:
            operators.append(next_operator)
            code = rem_code.strip()
            continue
        if next_operator is None and term_char is not None and code[0] != term_char and term_char in code:
            return None, rem_code

        # Check for term char
        if term_char is not None:
            if code[0] == term_char:
                break
            return None, f"Expected {term_char}"

        return None, rem_code
    if len(terms) == 1:
        return terms[0], code
    return Expression(terms, operators), code


def parse_assignment_statement(code: str) -> (AssignmentStatement, str):
    left, *more = code.split("=")
    if len(more) == 0:
        return None, "Expected = operator"
    if len(more) > 1:
        return None, "Expected only one = operator in assignment"
    right = more[0]
    left_expr, rem = parse_variable_expression(left.strip())
    if left_expr is None:
        return None, rem
    if len(rem) > 0:
        return None, "Expected only one variable expression for left term"
    right_expr, rem = parse_expression(right.strip())
    if right_expr is None:
        return None, rem
    return AssignmentStatement(left_expr, right_expr), None


class ForLoopParser(ABC):

    @abstractmethod
    def parse_for_loop(self, code: str) -> (ForLoopStatement, str):
        pass


class PseudocodeForLoopParser(ABC):

    HEADER_REGEX = re.compile(
        r"^FOR\s+(?P<idx>.+?)\s+FROM\s+(?P<lb>.+?)\s+TO\s+(?P<ub>.+?)(?:\s+STEP\s+(?P<step>.+?))?\s*\{\s*(?P<body>.*)\s*\}"
    )

    ASSIGNMENT_REGEX = re.compile(
        r"^STM\s+(?P<ass>.*?\s*=\s*.*?);.*"
    )

    def parse_loop_body(self, code: str) -> (tuple[Statement], str):
        code = code.strip()
        body_stmts = []
        while len(code) > 0:
            match = self.HEADER_REGEX.match(code)
            if match is not None:
                start, end = match.span()
                stmt, err = self.parse_for_loop(code[start:end])
                if stmt is None:
                    return None, err
                code = code[end:].strip()
                body_stmts.append(stmt)
                continue

            match = self.ASSIGNMENT_REGEX.match(code)
            if match is not None:
                assignment_code = match.group("ass").strip()
                stmt, err = parse_assignment_statement(assignment_code)
                if stmt is None:
                    return None, err
                end_of_assigment = match.span("ass")[1]+1
                code = code[end_of_assigment:].strip()
                body_stmts.append(stmt)
            else:
                return None, f"Unsupported statement: {code}"
        return body_stmts, None

    def parse_for_loop(self, code: str) -> (ForLoopStatement, str):
        code = re.sub(r"\s+", " ", code.strip())
        # Parse header
        if len(code) == 0:
            return None, "Expected loop header"
        match = self.HEADER_REGEX.match(code)
        if match is None:
            return None, "Incorrect loop header syntax"

        # Parse header parameters
        groups_dict = match.groupdict()
        idx_str = groups_dict["idx"]
        lb_str = groups_dict["lb"]
        ub_str = groups_dict["ub"]
        step_str = groups_dict["step"]
        idx_var_expr, err = parse_variable_expression(idx_str)
        if idx_var_expr is None:
            return None, err
        lb_expr, err = parse_expression(lb_str)
        if lb_expr is None:
            return None, err
        ub_expr, err = parse_expression(ub_str)
        if ub_expr is None:
            return None, err
        step_expr = None
        if step_str is not None:
            step_expr, err = parse_expression(step_str)
            if step_expr is None:
                return None, err

        # Parse body
        body_str = groups_dict["body"].strip()
        if body_str is None or len(body_str) == 0:
            return None, "Expected loop body"
        body, err = self.parse_loop_body(body_str)
        if body is None:
            return None, err

        if step_expr is not None:
            return ForLoopStatement(
                body=body,
                index=idx_var_expr,
                lowerbound=lb_expr,
                upperbound=ub_expr,
                step=step_expr
            ), None
        return ForLoopStatement(
            body=body,
            index=idx_var_expr,
            lowerbound=lb_expr,
            upperbound=ub_expr
        ), None



from abc import ABC, abstractmethod

from opoly.statements import Statement, ForLoopStatement, AssignmentStatement, divide_assignments, prune_expressions
from opoly.expressions import (
    Expression,
    ConstantExpression,
    VariableExpression,
    extract_variable_expressions,
    divide_variable_expressions_by_name
)


def is_simple_variable_sum_with_positive_constant(expr: Expression) -> bool:
    if (len(expr.operators) == 1 and
        (expr.operators[0] == "+" or expr.operators[0] == "-") and
        len(expr.terms) == 2
        ):
        if expr.terms[0].is_variable():
            return expr.terms[0].is_simple() and expr.terms[1].is_constant()
    return False

def get_simple_variable_sum_and_constant(expr: Expression) -> (VariableExpression, ConstantExpression):
    return expr.terms[0], ConstantExpression(expr.terms[1].value * (-1 if expr.operators[0] == "-" else 1))

def is_perfectly_nested_loop(loop: ForLoopStatement) -> bool:
    if len(loop.body) > 1:
        return not any([isinstance(st, ForLoopStatement) for st in loop.body])
    if isinstance(loop.body[0], ForLoopStatement):
        return is_perfectly_nested_loop(loop.body[0])
    return True


def get_inner_loop_statements(loop: ForLoopStatement) -> tuple[Statement]:
    if len(loop.body) == 1 and isinstance(loop.body[0], ForLoopStatement):
        return get_inner_loop_statements(loop.body[0])
    return loop.body


def is_plain_loop(loop: ForLoopStatement) -> bool:
    return ((loop.lowerbound.is_constant() and loop.lowerbound.value >= 0) and
            (((loop.upperbound.is_variable() and loop.upperbound.is_simple()) or
              is_simple_variable_sum_with_positive_constant(loop.upperbound))) and
            loop.step.is_constant() and loop.step.value == 1
            )


def is_recursively_plain_loop(loop: ForLoopStatement) -> bool:
    return is_plain_loop(loop) and all(
        map(
            lambda l: is_recursively_plain_loop(l),
            filter(lambda l: isinstance(l, ForLoopStatement), loop.body)
        )
    )


def get_indexed_variable_simple_indexes(var: VariableExpression) -> tuple[VariableExpression]:
    indexes = []
    for idx in var.indexes:
        if idx.is_variable() and idx.is_simple():
            indexes.append(idx)
        elif is_simple_variable_sum_with_positive_constant(idx):
            indexes.append(idx.terms[0])
        else:
            return None
    return indexes


def get_indexed_variable_index_constants(var: VariableExpression) -> tuple[ConstantExpression]:
    constants = []
    for idx in var.indexes:
        if idx.is_variable() and idx.is_simple():
            constants.append(ConstantExpression(0))
        elif is_simple_variable_sum_with_positive_constant(idx):
            constants.append(idx.terms[1])
        else:
            return None
    return constants


def get_indexed_variable_index_constants_with_sign(var: VariableExpression) -> tuple[ConstantExpression]:
    constants = []
    for idx in var.indexes:
        if idx.is_variable() and idx.is_simple():
            constants.append(ConstantExpression(0))
        elif is_simple_variable_sum_with_positive_constant(idx):
            constants.append(ConstantExpression(
                    idx.terms[1].value * (-1 if idx.operators[0] == "-" else 1)))
        else:
            return None
    return constants


def extract_loop_indexes(loop: ForLoopStatement) -> tuple[VariableExpression]:
    inner_indexes = []
    if isinstance(loop.body[0], ForLoopStatement):
        inner_indexes = extract_loop_indexes(loop.body[0])
    return tuple([loop.index] + list(inner_indexes))

def extract_loop_bounds(loop: ForLoopStatement) -> tuple[(Expression, Expression)]:
    inner_bounds = []
    if isinstance(loop.body[0], ForLoopStatement):
        inner_bounds = extract_loop_bounds(loop.body[0])
    return tuple([(loop.lowerbound, loop.upperbound)] + list(inner_bounds))


def has_same_simple_indexes(var1: VariableExpression, var2: VariableExpression) -> bool:
    var1_indexes = get_indexed_variable_simple_indexes(var1)
    var2_indexes = get_indexed_variable_simple_indexes(var2)
    return var1_indexes is not None and var2_indexes is not None and \
        list(idx.name for idx in var1_indexes) == list(
            idx.name for idx in var2_indexes)


class ForLoopChecker(ABC):

    def __init__(self, **params):
        self._params_dict = params

    @abstractmethod
    def check(self, loop: ForLoopStatement) -> (bool, str):
        pass


class LamportForLoopChecker(ForLoopChecker):

    def check(self, loop: ForLoopStatement) -> (bool, str):
        if not is_perfectly_nested_loop(loop):
            return False, "Not perfectly nested loop!"
        if not is_recursively_plain_loop(loop):
            return False, "Not plain loop!"
        indexes = extract_loop_indexes(loop)
        if not all(map(lambda i: i.is_simple(), indexes)):
            return False, "Loop indexes are not simple!"
        index_names = tuple([i.name for i in indexes])  # pylint: disable=no-member
        if not len(set(index_names)) == len(index_names):
            return False, "Not all index names in loop are distinct!"
        inner_statements = get_inner_loop_statements(loop)
        if not all(map(lambda stmt: isinstance(stmt, AssignmentStatement), inner_statements)):
            return False, "Not all statements in loop body are assignments!"
        lefts, rights = divide_assignments(inner_statements)
        if not all(map(lambda l: l.is_variable(), lefts)):
            return False, "Not all left sides of statements in loop body are variable expressions!"
        generations, uses = prune_expressions(lefts, rights)
        if any(map(lambda gen: gen.is_simple(), generations)):
            return False, "All variable generations must be non-simple!"
        all_variables = tuple(list(generations) + list(uses))
        all_variables_by_name = divide_variable_expressions_by_name(
            all_variables)
        for _, same_name_vars in all_variables_by_name.items():
            all_simple = all(map(lambda v: v.is_simple(), same_name_vars))
            all_non_simple = all(
                map(lambda v: not v.is_simple(), same_name_vars))
            if not (all_simple or all_non_simple):
                return False, "Same name variables must be all simple or all non-simple!"
        # Filter non-simple uses
        uses = tuple(filter(lambda use: not use.is_simple(), uses))
        all_variables = tuple(list(generations) + list(uses))

        # Check variable indexes
        var_names_dict = divide_variable_expressions_by_name(all_variables)
        for _, same_name_vars in var_names_dict.items():
            for var1 in same_name_vars:
                var1_indexes = get_indexed_variable_simple_indexes(var1)
                var1_constants = get_indexed_variable_index_constants(var1)
                if var1_indexes is None or var1_constants is None:
                    return False, f"Variable expression ({str(var1)}) has incorrect indexes!"
                var1_index_names = tuple([i.name for i in var1_indexes])

                if not all(map(lambda i: i.name in index_names, var1_indexes)):
                    return False, f"Variable expression ({str(var1)}) has an index not present in loop indexes!"
                if not len(set(var1_index_names)) == len(var1_index_names):
                    return False, f"Variable expression ({str(var1)}) indexes are not unique!"
                if not all(map(lambda iname: iname in var1_index_names, index_names[1:])):
                    return False, f"Variable expression ({str(var1)}) has a missing index which is not ({index_names[0]})!"
                if not all(map(lambda con: isinstance(con.value, int), var1_constants)):
                    return False, f"Variable expression ({str(var1)}) has a non-integer constant in one of the indexes!"

            for var1 in same_name_vars:
                for var2 in same_name_vars:
                    if not has_same_simple_indexes(var1, var2):
                        return False, f"Variable expressions ({str(var1)}) and ({str(var2)}) have the same name but different indexes order!"
        return (True, None)



from pkg_resources import resource_filename

import pymzn

import opoly.modules.minizinc as minizinc


def solve_model(
    model: str,
    data: dict,
    solver=pymzn.chuffed,
    timeout: int = 5
) -> (pymzn.Solutions, str):
    try:
        sols = pymzn.minizinc(
            mzn=model,
            data=data,
            solver=solver,
            timeout=timeout,
            include=minizinc.INCLUDE_FOLDER_PATH
        )
    except Exception as ex:
        return None, f"An error occurred!\n{ex}"
    if sols.status in (pymzn.Status.UNKNOWN, pymzn.Status.INCOMPLETE):
        return None, "Solution not found in time!"
    if sols.status == pymzn.Status.UNSATISFIABLE:
        return None, "Unsatisfiable!"
    return sols[0], None
